# Copyright (c) Facebook, Inc. and its affiliates.
# Copyright (c) 2021 Dhruv Agarwal and authors of arboEL.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import os
import json
import random
import math
import time
import torch
import numpy as np
from tqdm import tqdm
import pickle
import copy
from sklearn.cluster import KMeans

import bioel.models.arboel.data.data_process as data_process
from bioel.models.arboel.model.eval_cluster_linking import (
    analyzeClusters,
    partition_graph,
)

# from blink.biencoder.eval_entity_discovery import (
#     analyzeClusters as analyze_discovery_clusters,
# )
from bioel.models.arboel.model.biencoder import BiEncoderRanker
from bioel.models.arboel.crossencoder.original.crossencoder import CrossEncoderRanker
from bioel.models.arboel.model.common.params import BlinkParser
from bioel.models.arboel.crossencoder.train_crossencoder_mst import (
    get_context_doc_ids,
    get_biencoder_nns,
    build_cross_concat_input,
    score_in_batches,
)
from bioel.models.arboel.crossencoder.original.train_cross import read_dataset

from IPython import embed

from bioel.logger import setup_logger

SCORING_BATCH_SIZE = 64


def load_data(
    data_split,
    bi_tokenizer,
    max_context_length,
    max_cand_length,
    knn,
    pickle_src_path,
    params,
    logger,
    return_dict_only=False,
):
    entity_dictionary_loaded = False
    entity_dictionary_pkl_path = os.path.join(
        pickle_src_path, "entity_dictionary.pickle"
    )
    if os.path.isfile(entity_dictionary_pkl_path):
        print("Loading stored processed entity dictionary...")
        with open(entity_dictionary_pkl_path, "rb") as read_handle:
            entity_dictionary = pickle.load(read_handle)
        entity_dictionary_loaded = True

    if return_dict_only and entity_dictionary_loaded:
        return entity_dictionary

    # Load data
    tensor_data_pkl_path = os.path.join(
        pickle_src_path, f"{data_split}_tensor_data.pickle"
    )
    processed_data_pkl_path = os.path.join(
        pickle_src_path, f"{data_split}_processed_data.pickle"
    )
    if os.path.isfile(tensor_data_pkl_path) and os.path.isfile(processed_data_pkl_path):
        print("Loading stored processed data...")
        with open(tensor_data_pkl_path, "rb") as read_handle:
            tensor_data = pickle.load(read_handle)
        with open(processed_data_pkl_path, "rb") as read_handle:
            processed_data = pickle.load(read_handle)
    else:
        data_samples = read_dataset(data_split, params["data_path"])
        if not entity_dictionary_loaded:
            with open(
                os.path.join(params["data_path"], "dictionary.pickle"), "rb"
            ) as read_handle:
                entity_dictionary = pickle.load(read_handle)

        # Check if dataset has multiple ground-truth labels
        mult_labels = "labels" in data_samples[0].keys()
        # Filter samples without gold entities
        data_samples = list(
            filter(
                lambda sample: (
                    (len(sample["labels"]) > 0)
                    if mult_labels
                    else (sample["label"] is not None)
                ),
                data_samples,
            )
        )
        logger.info("Read %d data samples." % len(data_samples))

        processed_data, entity_dictionary, tensor_data = (
            data_process.process_mention_data(
                data_samples,
                entity_dictionary,
                bi_tokenizer,
                max_context_length,
                max_cand_length,
                context_key=params["context_key"],
                multi_label_key="labels" if mult_labels else None,
                silent=params["silent"],
                logger=logger,
                debug=params["debug"],
                knn=knn,
                dictionary_processed=entity_dictionary_loaded,
            )
        )
        print("Saving processed data...")
        if not entity_dictionary_loaded:
            with open(entity_dictionary_pkl_path, "wb") as write_handle:
                pickle.dump(
                    entity_dictionary, write_handle, protocol=pickle.HIGHEST_PROTOCOL
                )
        with open(tensor_data_pkl_path, "wb") as write_handle:
            pickle.dump(tensor_data, write_handle, protocol=pickle.HIGHEST_PROTOCOL)
        with open(processed_data_pkl_path, "wb") as write_handle:
            pickle.dump(processed_data, write_handle, protocol=pickle.HIGHEST_PROTOCOL)

    if return_dict_only:
        return entity_dictionary
    return entity_dictionary, tensor_data, processed_data


def filter_by_context_doc_id(mention_idxs, doc_id, doc_id_list, return_numpy=False):
    mask = [doc_id_list[i] == doc_id for i in mention_idxs]
    if isinstance(mention_idxs, list):
        mention_idxs = np.array(mention_idxs)
    mention_idxs = mention_idxs[mask]
    if not return_numpy:
        mention_idxs = list(mention_idxs)
    return mention_idxs, mask


def get_entity_idxs_to_drop(processed_data, params, logger):
    print("******** Discovery ********")
    # Proportion of unique entities in the mention set to drop
    ent_drop_prop = params["ent_drop_prop"]
    mention_gold_entities = list(map(lambda x: x["label_idxs"][0], processed_data))
    ents_in_data = np.unique(mention_gold_entities)
    logger.info(
        f"Dropping {ent_drop_prop * 100}% of {len(ents_in_data)} entities found in mention set"
    )

    # Get entity idxs to drop
    n_ents_dropped = int(ent_drop_prop * len(ents_in_data))
    rng = np.random.default_rng(seed=17)  # Random number generator
    dropped_ent_idxs = rng.choice(ents_in_data, size=n_ents_dropped, replace=False)
    set_dropped_ent_idxs = set(dropped_ent_idxs)

    # Calculate number of mentions without gold entities after dropping
    n_mentions_wo_gold_ents = sum(
        [1 if x in set_dropped_ent_idxs else 0 for x in mention_gold_entities]
    )
    logger.info(f"Dropped {n_ents_dropped} entities")
    logger.info(f"=> Mentions without gold entities = {n_mentions_wo_gold_ents}")
    print("*****************")
    return (
        set_dropped_ent_idxs,
        n_ents_dropped,
        n_mentions_wo_gold_ents,
        mention_gold_entities,
    )


def run_inference(
    entity_dictionary,
    processed_data,
    results,
    result_overview,
    joint_graphs,
    n_entities,
    time_start,
    output_path,
    bi_recall,
    k_biencoder,
    logger,
):
    knn_fetch_time = time.time() - time_start
    n_graphs_processed = 0
    graph_processing_time = time.time()
    for mode in results:
        print(f"\nEvaluation mode: {mode.upper()}")
        for k in joint_graphs:
            if k == 0 and mode == "undirected" and len(results) > 1:
                continue  # Since @k=0 both modes are equivalent, so skip for one mode
            logger.info(f"\nGraph (k={k}):")
            # Partition graph based on cluster-linking constraints
            partitioned_graph, clusters = partition_graph(
                joint_graphs[k],
                n_entities,
                directed=(mode == "directed"),
                return_clusters=True,
            )
            # Infer predictions from clusters
            result = analyzeClusters(clusters, entity_dictionary, processed_data, k)
            # Store result
            results[mode].append(result)
            n_graphs_processed += 1

    avg_graph_processing_time = (
        time.time() - graph_processing_time
    ) / n_graphs_processed
    avg_per_graph_time = (knn_fetch_time + avg_graph_processing_time) / 60

    execution_time = (time.time() - time_start) / 60
    # Store results
    output_file_name = os.path.join(
        output_path,
        f"eval_results_{__import__('calendar').timegm(__import__('time').gmtime())}",
    )

    if bi_recall is not None:
        result_overview[f"biencoder recall@{k_biencoder}"] = f"{bi_recall * 100} %"
    else:
        logger.info("Recall data not available (graphs were loaded from disk)")

    for mode in results:
        mode_results = results[mode]
        result_overview[mode] = {}
        for r in mode_results:
            k = r["knn_mentions"]
            result_overview[mode][f"accuracy@knn{k}"] = r["accuracy"]
            logger.info(f"{mode} accuracy@knn{k} = {r['accuracy']}")
            output_file = f"{output_file_name}-{mode}-{k}.json"
            with open(output_file, "w") as f:
                json.dump(r, f, indent=2)
                print(f"\nPredictions ({mode}) @knn{k} saved at: {output_file}")
    with open(f"{output_file_name}.json", "w") as f:
        json.dump(result_overview, f, indent=2)
        print(f"\nPredictions overview saved at: {output_file_name}.json")

    logger.info(
        "\nThe avg. per graph evaluation time is {} minutes\n".format(
            avg_per_graph_time
        )
    )
    logger.info("\nThe total evaluation took {} minutes\n".format(execution_time))


def run_discovery_experiment(
    joint_graphs,
    dropped_ent_idxs,
    mention_gold_entities,
    n_entities,
    n_mentions,
    data_split,
    results,
    output_path,
    time_start,
    params,
    logger,
):
    def discovery_helper(
        results, best_result=None, best_config=None, drop_all_entities=False
    ):
        drop_set = dropped_ent_idxs
        if drop_all_entities:
            drop_set = set([i for i in range(n_entities)])
        if exact_threshold is not None:
            thresholds = np.array([0, exact_threshold])
        else:
            thresholds = np.sort(
                np.concatenate(
                    (
                        [0],
                        k_means.fit(
                            joint_graphs[k]["data"].reshape(-1, 1)
                        ).cluster_centers_.flatten(),
                    )
                )
            )
        for thresh in thresholds:
            print()
            logger.info("Partitioning...")
            logger.info(f"{mode.upper()}, k={k}, threshold={thresh}")
            # Partition graph based on cluster-linking constraints
            partitioned_graph, clusters = partition_graph(
                joint_graphs[k],
                n_entities,
                directed=(mode == "directed"),
                return_clusters=True,
                exclude=drop_set,
                threshold=thresh,
            )
            # Analyze cluster against gold clusters
            result = analyze_discovery_clusters(
                clusters, mention_gold_entities, n_entities, n_mentions, logger
            )
            results[f"({mode}, {k}, {thresh})"] = result
            if best_result is not None:
                if thresh != 0 and result["average"] > best_result:
                    best_result = result["average"]
                    best_config = (mode, k, thresh)
        return best_result, best_config

    graph_mode = ["directed", "undirected"]
    # Number of similarity thresholds to try in order to find best clustering; Default=10
    n_thresholds = params["n_thresholds"]
    k_means = KMeans(n_clusters=n_thresholds, random_state=17)
    # Specific threshold to run the experiment with, if the param is passed
    exact_threshold = params.get("exact_threshold", None)
    # Specific value of K defining the number of mention nearest-neighbors used for clustering, if the param is passed
    exact_knn = params.get("exact_knn", None)
    # Store the baseline results
    baselines = {}

    for mode in graph_mode:
        best_result = -1.0
        best_config = None
        for k in joint_graphs:
            if k == 0:
                continue
            # First run the baseline (i.e. with no entity edges), which stores results in the passed arg
            logger.info("Baseline:")
            discovery_helper(baselines, drop_all_entities=True)
            if (exact_knn is None and 0 < k <= params["knn"]) or (
                exact_knn is not None and k == exact_knn
            ):
                best_result, best_config = discovery_helper(
                    results, best_result, best_config
                )
        results[f"best_{mode}_config"] = best_config
        results[f"best_{mode}_result"] = best_result
        results["baselines"] = baselines

    # Store results
    output_file_name = os.path.join(
        output_path,
        f"{data_split}_eval_discovery_{__import__('calendar').timegm(__import__('time').gmtime())}.json",
    )

    with open(output_file_name, "w") as f:
        json.dump(results, f, indent=2)
        print(f"\nAnalysis saved at: {output_file_name}")
    execution_time = (time.time() - time_start) / 60
    logger.info(f"\nTotal time taken: {execution_time} minutes\n")


def save_topk_biencoder_cands(
    bi_reranker,
    use_types,
    logger,
    n_gpu,
    params,
    bi_tokenizer,
    max_context_length,
    max_cand_length,
    pickle_src_path,
    topk=64,
):
    entity_dictionary = load_data(
        "train",
        bi_tokenizer,
        max_context_length,
        max_cand_length,
        1,
        pickle_src_path,
        params,
        logger,
        return_dict_only=True,
    )
    entity_dict_vecs = torch.tensor(
        list(map(lambda x: x["ids"], entity_dictionary)), dtype=torch.long
    )

    logger.info("Biencoder: Embedding and indexing entity dictionary")
    if use_types:
        _, dict_indexes, dict_idxs_by_type = data_process.embed_and_index(
            bi_reranker,
            entity_dict_vecs,
            encoder_type="candidate",
            corpus=entity_dictionary,
            force_exact_search=True,
            batch_size=params["embed_batch_size"],
        )
    else:
        _, dict_index = data_process.embed_and_index(
            bi_reranker,
            entity_dict_vecs,
            encoder_type="candidate",
            force_exact_search=True,
            batch_size=params["embed_batch_size"],
        )
    logger.info("Biencoder: Embedding and indexing finished")

    for mode in ["train", "valid", "test"]:
        logger.info(
            f"Biencoder: Fetching top-{topk} biencoder candidates for {mode} set"
        )
        _, tensor_data, processed_data = load_data(
            mode,
            bi_tokenizer,
            max_context_length,
            max_cand_length,
            1,
            pickle_src_path,
            params,
            logger,
        )
        men_vecs = tensor_data[:][0]

        logger.info("Biencoder: Embedding mention data")
        if use_types:
            men_embeddings, _, men_idxs_by_type = data_process.embed_and_index(
                bi_reranker,
                men_vecs,
                encoder_type="context",
                corpus=processed_data,
                force_exact_search=True,
                batch_size=params["embed_batch_size"],
            )
        else:
            men_embeddings = data_process.embed_and_index(
                bi_reranker,
                men_vecs,
                encoder_type="context",
                force_exact_search=True,
                batch_size=params["embed_batch_size"],
                only_embed=True,
            )
        logger.info("Biencoder: Embedding finished")

        logger.info("Biencoder: Finding nearest entities for each mention...")
        if not use_types:
            _, bi_dict_nns = dict_index.search(men_embeddings, topk)
        else:
            bi_dict_nns = np.zeros((len(men_embeddings), topk), dtype=int)
            for entity_type in men_idxs_by_type:
                men_embeds_by_type = men_embeddings[men_idxs_by_type[entity_type]]
                _, dict_nns_by_type = dict_indexes[entity_type].search(
                    men_embeds_by_type, topk
                )
                dict_nns_idxs = np.array(
                    list(
                        map(
                            lambda x: dict_idxs_by_type[entity_type][x],
                            dict_nns_by_type,
                        )
                    )
                )
                for i, idx in enumerate(men_idxs_by_type[entity_type]):
                    bi_dict_nns[idx] = dict_nns_idxs[i]
        logger.info("Biencoder: Search finished")

        labels = [-1] * len(bi_dict_nns)
        for men_idx in range(len(bi_dict_nns)):
            gold_idx = processed_data[men_idx]["label_idxs"][0]
            for i in range(len(bi_dict_nns[men_idx])):
                if bi_dict_nns[men_idx][i] == gold_idx:
                    labels[men_idx] = i
                    break

        logger.info(f"Biencoder: Saving top-{topk} biencoder candidates for {mode} set")
        save_data_path = os.path.join(
            params["output_path"], f"candidates_{mode}_top{topk}.t7"
        )
        torch.save(
            {"mode": mode, "candidates": bi_dict_nns, "labels": labels}, save_data_path
        )
        logger.info("Biencoder: Saved")

    return {"mode": mode, "candidates": bi_dict_nns, "labels": labels}


def main(params):
    # Parameter initializations
    logger = setup_logger()
    global SCORING_BATCH_SIZE
    SCORING_BATCH_SIZE = params["scoring_batch_size"]
    output_path = params["output_path"]
    if not os.path.exists(output_path):
        os.makedirs(output_path)
    pickle_src_path = params["pickle_src_path"]
    if pickle_src_path is None or not os.path.exists(pickle_src_path):
        pickle_src_path = output_path
    biencoder_indices_path = params["biencoder_indices_path"]
    if biencoder_indices_path is None:
        biencoder_indices_path = output_path
    elif not os.path.exists(biencoder_indices_path):
        os.makedirs(biencoder_indices_path)
    max_k = params["knn"]  # Maximum k-NN graph to build for evaluation
    use_types = params["use_types"]
    within_doc = params["within_doc"]
    discovery_mode = params["discovery"]

    # Bi-encoder model
    biencoder_params = copy.deepcopy(params)
    biencoder_params["add_linear"] = False
    bi_reranker = BiEncoderRanker(biencoder_params)
    bi_tokenizer = bi_reranker.tokenizer
    k_biencoder = params[
        "bi_knn"
    ]  # Number of biencoder nearest-neighbors to fetch for cross-encoder scoring (default: 64)

    # Cross-encoder model
    params["add_linear"] = True
    params["add_sigmoid"] = True
    cross_reranker = CrossEncoderRanker(params)
    n_gpu = cross_reranker.n_gpu
    cross_reranker.model.eval()

    # Input lengths
    max_seq_length = params["max_seq_length"]
    max_context_length = params["max_context_length"]
    max_cand_length = params["max_cand_length"]

    # Fix random seeds
    seed = params["seed"]
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if cross_reranker.n_gpu > 0:
        torch.cuda.manual_seed_all(seed)

    # The below code is to generate the candidates for cross-encoder training and inference
    if params["save_topk_result"]:
        save_topk_biencoder_cands(
            bi_reranker,
            use_types,
            logger,
            n_gpu,
            params,
            bi_tokenizer,
            max_context_length,
            max_cand_length,
            pickle_src_path,
            topk=64,
        )
        exit()

    data_split = params["data_split"]
    entity_dictionary, tensor_data, processed_data = load_data(
        data_split,
        bi_tokenizer,
        max_context_length,
        max_cand_length,
        max_k,
        pickle_src_path,
        params,
        logger,
    )
    n_entities = len(entity_dictionary)
    n_mentions = len(processed_data)
    # Store dictionary vectors
    dict_vecs = torch.tensor(
        list(map(lambda x: x["ids"], entity_dictionary)), dtype=torch.long
    )
    # Store query vectors
    men_vecs = tensor_data[:][0]

    discovery_entities = []
    if discovery_mode:
        (
            discovery_entities,
            n_ents_dropped,
            n_mentions_wo_gold_ents,
            mention_gold_entities,
        ) = get_entity_idxs_to_drop(processed_data, params, logger)

    context_doc_ids = None
    if within_doc:
        # Get context_document_ids for each mention in training and validation
        context_doc_ids = get_context_doc_ids(data_split, params)
    params["only_evaluate"] = True  # Needed to call get_biencoder_nns() correctly
    _, biencoder_nns = get_biencoder_nns(
        bi_reranker=bi_reranker,
        biencoder_indices_path=biencoder_indices_path,
        entity_dictionary=entity_dictionary,
        entity_dict_vecs=dict_vecs,
        train_men_vecs=None,
        train_processed_data=None,
        train_gold_clusters=None,
        valid_men_vecs=men_vecs,
        valid_processed_data=processed_data,
        use_types=use_types,
        logger=logger,
        n_gpu=n_gpu,
        params=params,
        train_context_doc_ids=None,
        valid_context_doc_ids=context_doc_ids,
    )
    bi_men_idxs = biencoder_nns["men_nns"][:, :k_biencoder]
    bi_ent_idxs = biencoder_nns["dict_nns"][:, :k_biencoder]
    bi_nn_count = np.sum(biencoder_nns["men_nns"] != -1, axis=1)

    # Compute and store the concatenated cross-encoder inputs for validation
    men_concat_inputs, ent_concat_inputs = build_cross_concat_input(
        biencoder_nns, men_vecs, dict_vecs, max_seq_length, k_biencoder
    )

    # Values of k to run the evaluation against
    knn_vals = [0] + [2**i for i in range(int(math.log(max_k, 2)) + 1)]
    max_k = knn_vals[-1]  # Store the maximum evaluation k
    bi_recall = None

    time_start = time.time()
    # Check if k-NN graphs are already built
    graph_path = os.path.join(output_path, "graphs.pickle")
    if not params["only_recall"] and os.path.isfile(graph_path):
        print("Loading stored joint graphs...")
        with open(graph_path, "rb") as read_handle:
            joint_graphs = pickle.load(read_handle)
    else:
        joint_graphs = {}
        for k in knn_vals:
            joint_graphs[k] = {
                "rows": np.array([]),
                "cols": np.array([]),
                "data": np.array([]),
                "shape": (n_entities + n_mentions, n_entities + n_mentions),
            }

        # Score biencoder NNs using cross-encoder
        score_path = os.path.join(output_path, "cross_scores_indexes.pickle")
        if os.path.isfile(score_path):
            print("Loading stored cross-encoder scores and indexes...")
            with open(score_path, "rb") as read_handle:
                score_data = pickle.load(read_handle)
            cross_men_topk_idxs = score_data["cross_men_topk_idxs"]
            cross_men_topk_scores = score_data["cross_men_topk_scores"]
            cross_ent_top1_idx = score_data["cross_ent_top1_idx"]
            cross_ent_top1_score = score_data["cross_ent_top1_score"]
        else:
            with torch.no_grad():
                logger.info(
                    "Eval: Scoring mention-mention edges using cross-encoder..."
                )
                cross_men_scores = score_in_batches(
                    cross_reranker,
                    max_context_length,
                    men_concat_inputs,
                    is_context_encoder=True,
                    scoring_batch_size=SCORING_BATCH_SIZE,
                )
                for i in range(len(cross_men_scores)):
                    # Set scores for all invalid nearest neighbours to -infinity (due to variable NN counts of mentions)
                    cross_men_scores[i][bi_nn_count[i] :] = float("-inf")
                cross_men_topk_scores, cross_men_topk_idxs = torch.sort(
                    cross_men_scores, dim=1, descending=True
                )
                cross_men_topk_idxs = cross_men_topk_idxs.cpu()[:, :max_k]
                cross_men_topk_scores = cross_men_topk_scores.cpu()[:, :max_k]
                logger.info("Eval: Scoring done")

                logger.info("Eval: Scoring mention-entity edges using cross-encoder...")
                cross_ent_scores = score_in_batches(
                    cross_reranker,
                    max_context_length,
                    ent_concat_inputs,
                    is_context_encoder=False,
                    scoring_batch_size=SCORING_BATCH_SIZE,
                )
                cross_ent_top1_score, cross_ent_top1_idx = torch.sort(
                    cross_ent_scores, dim=1, descending=True
                )
                cross_ent_top1_idx = cross_ent_top1_idx.cpu()
                cross_ent_top1_score = cross_ent_top1_score.cpu()
                if discovery_mode:
                    # Replace the first value in each row with an entity not in the drop set
                    for i in range(cross_ent_top1_idx.shape[0]):
                        for j in range(cross_ent_top1_idx.shape[1]):
                            if cross_ent_top1_idx[i, j] not in discovery_entities:
                                cross_ent_top1_idx[i, 0] = cross_ent_top1_idx[i, j]
                                cross_ent_top1_score[i, 0] = cross_ent_top1_score[i, j]
                                break
                cross_ent_top1_idx = cross_ent_top1_idx[:, 0]
                cross_ent_top1_score = cross_ent_top1_score[:, 0]
                logger.info("Eval: Scoring done")
            # Pickle the scores and nearest indexes
            logger.info("Saving cross-encoder scores and indexes...")
            with open(score_path, "wb") as write_handle:
                pickle.dump(
                    {
                        "cross_men_topk_idxs": cross_men_topk_idxs,
                        "cross_men_topk_scores": cross_men_topk_scores,
                        "cross_ent_top1_idx": cross_ent_top1_idx,
                        "cross_ent_top1_score": cross_ent_top1_score,
                    },
                    write_handle,
                    protocol=pickle.HIGHEST_PROTOCOL,
                )
            logger.info(f"Saved at: {score_path}")

        # Build k-NN graphs
        bi_recall = 0.0
        for men_idx in tqdm(
            range(len(processed_data)),
            total=len(processed_data),
            desc="Eval: Building graphs",
        ):
            # Track biencoder recall@<k_biencoder>
            gold_idx = processed_data[men_idx]["label_idxs"][0]
            if gold_idx in bi_ent_idxs[men_idx]:
                bi_recall += 1.0
            # Get nearest entity
            m_e_idx = bi_ent_idxs[men_idx, cross_ent_top1_idx[men_idx]]
            m_e_score = cross_ent_top1_score[men_idx]
            if bi_nn_count[men_idx] > 0:
                # Get nearest mentions
                topk_defined_nn_idxs = cross_men_topk_idxs[men_idx][
                    : bi_nn_count[men_idx]
                ]
                m_m_idxs = (
                    bi_men_idxs[men_idx, topk_defined_nn_idxs] + n_entities
                )  # Mentions added at an offset of maximum entities
                m_m_scores = cross_men_topk_scores[men_idx][: bi_nn_count[men_idx]]
            # Add edges to the graphs
            for k in joint_graphs:
                # Add mention-entity edge
                joint_graphs[k]["rows"] = np.append(
                    joint_graphs[k]["rows"], [n_entities + men_idx]
                )  # Mentions added at an offset of maximum entities
                joint_graphs[k]["cols"] = np.append(joint_graphs[k]["cols"], m_e_idx)
                joint_graphs[k]["data"] = np.append(joint_graphs[k]["data"], m_e_score)
                if k > 0 and bi_nn_count[men_idx] > 0:
                    # Add mention-mention edges
                    joint_graphs[k]["rows"] = np.append(
                        joint_graphs[k]["rows"],
                        [n_entities + men_idx] * len(m_m_idxs[:k]),
                    )
                    joint_graphs[k]["cols"] = np.append(
                        joint_graphs[k]["cols"], m_m_idxs[:k]
                    )
                    joint_graphs[k]["data"] = np.append(
                        joint_graphs[k]["data"], m_m_scores[:k]
                    )
        # Pickle the graphs
        print("Saving joint graphs...")
        with open(graph_path, "wb") as write_handle:
            pickle.dump(joint_graphs, write_handle, protocol=pickle.HIGHEST_PROTOCOL)
        logger.info(f"Saved graphs at: {graph_path}")
        # Compute biencoder recall
        bi_recall /= len(processed_data)
        if params["only_recall"]:
            logger.info(f"Eval: Biencoder recall@{k_biencoder} = {bi_recall * 100}%")
            exit()

    graph_mode = params.get("graph_mode", None)

    if discovery_mode:
        # Run the entity discovery experiment
        results = {
            "data_split": data_split.upper(),
            "n_entities": n_entities,
            "n_mentions": n_mentions,
            "n_entities_dropped": f"{n_ents_dropped} ({params['ent_drop_prop'] * 100}%)",
            "n_mentions_wo_gold_entities": n_mentions_wo_gold_ents,
        }
        run_discovery_experiment(
            joint_graphs,
            discovery_entities,
            mention_gold_entities,
            n_entities,
            n_mentions,
            data_split,
            results,
            output_path,
            time_start,
            params,
            logger,
        )
    else:
        # Run entity linking inference
        result_overview, results = {
            "n_entities": n_entities,
            "n_mentions": n_mentions,
        }, {}
        if graph_mode is None or graph_mode not in ["directed", "undirected"]:
            results["directed"], results["undirected"] = [], []
        else:
            results[graph_mode] = []
        run_inference(
            entity_dictionary,
            processed_data,
            results,
            result_overview,
            joint_graphs,
            n_entities,
            time_start,
            output_path,
            bi_recall,
            k_biencoder,
            logger,
        )


if __name__ == "__main__":
    parser = BlinkParser(add_model_args=True)
    parser.add_eval_args()
    parser.add_joint_train_args()
    args = parser.parse_args()
    print(args)
    main(args.__dict__)



import torch
import numpy as np
from tqdm import tqdm
from torch.utils.data import TensorDataset
from torch.utils.data import DataLoader, SequentialSampler, DistributedSampler
from torch.distributed import all_gather, get_rank, broadcast
import torch.distributed as dist
import math
import faiss
import os
from IPython import embed

from bioel.models.arboel.biencoder.model.common.params import (
    ENT_START_TAG,
    ENT_END_TAG,
    ENT_TITLE_TAG,
)


def select_field(data, key1, key2=None):
    """
    Description
    -----------
    #J Extracts specific data from a LIST of dictionaries (key1) or a list of nested dictionaries (key2)

    Parameters
    ----------
    data :
        List of dictionaries
    key1 :
        First key to look up in each dictionary
    key2 :
        Optional second key
    """
    if key2 is None:
        return [example[key1] for example in data]
    else:
        return [example[key1][key2] for example in data]


def get_context_representation(
    sample,
    tokenizer,
    max_seq_length,
    mention_key="mention",
    context_key="context",
    ent_start_token=ENT_START_TAG,
    ent_end_token=ENT_END_TAG,
):
    """
    Description
    -----------
    Prepare the text data for input into a transformer-based model like BERT.
    Given a sample of text, it uses a tokenizer to convert the text into a format that is compatible with these models,
    which typically require input to be in a specific tokenized format with special tokens and padding.

    Parameters
    ----------
    sample : #D1 list of dics
        Each dict represents a text sample with mention and context. (And we have plenty of them)
    tokenizer : Object
        Def : Converts text into token (smaller chunk of text such as subwords)
    max_seq_length : int
        Maximum length of the token sequence for the model input.
    mention_key : str
        Key in the sample dictionary that holds the text of the mention
    context_key : str
        Key used to access the surrounding context of the mention in the sample dictionary.
    ent_start_token :
        Token before the mention to indicate its start.
    ent_end_token :
        Token that is added after the mention to indicate its end

    -------
    Returns {tokens + ID} of "[CLS] c_left [START] mention [END] c_right [SEP]"
    """
    mention_tokens = []
    if (
        sample[mention_key] and len(sample[mention_key]) > 0
    ):  # redundant : one condition would have been enough
        mention_tokens = tokenizer.tokenize(sample[mention_key])
        mention_tokens = [ent_start_token] + mention_tokens + [ent_end_token]

    # Tokenization of context
    context_left = sample[context_key + "_left"]  # = sample["context_left"]
    context_left = tokenizer.tokenize(context_left)  # tokenize the sentences
    context_right = sample[context_key + "_right"]  # = sample["context_right"]
    context_right = tokenizer.tokenize(context_right)

    # D2 Quota Calculation for left and right context
    left_quota = (max_seq_length - len(mention_tokens)) // 2 - 1
    right_quota = max_seq_length - len(mention_tokens) - left_quota - 2
    left_add = len(context_left)
    right_add = len(context_right)
    if left_add <= left_quota:
        if right_add > right_quota:
            right_quota += left_quota - left_add
    else:
        if right_add <= right_quota:
            left_quota += right_quota - right_add

    context_tokens = (
        context_left[-left_quota:] + mention_tokens + context_right[:right_quota]
    )
    context_tokens = ["[CLS]"] + context_tokens + ["[SEP]"]

    # Conversion to IDs and Padding
    # The context tokens are converted to IDs, and then padding is added to ensure the sequence length is equal to max_seq_length.
    input_ids = tokenizer.convert_tokens_to_ids(context_tokens)
    padding = [0] * (max_seq_length - len(input_ids))
    input_ids += padding
    assert len(input_ids) == max_seq_length

    # Output : Returns the processed tokens and their corresponding IDs in a dictionary.
    return {
        "tokens": context_tokens,
        "ids": input_ids,
    }


def get_candidate_representation(
    candidate_desc,
    tokenizer,
    max_seq_length,
    candidate_title=None,
    title_tag=ENT_TITLE_TAG,
):
    """
    Description
    -----------
    Take a description of a candidate entity and convert it into a format suitable for input into a transformer-based model (= tokens + ID).

    Parameters
    ----------
    candidate_desc: str
        Description of the candidate entity
    tokenizer : Object
        Def : Converts text into token (smaller chunk of text such as subwords)
    max_seq_length : int
        Maximum length of the token sequence for the model input.
    candidate_title : str, optional
        Title of the candidate entity.
    title_tag : str
        Special token added to the token sequence to separate the candidate_title from the candidate_desc
    -------
    Returns {tokens + ID} of "[CLS] e_title [TITLE] e_desc [SEP]"
    """
    # Retrieves the special [CLS] and [SEP] tokens used by the BERT tokenizer (can also be from other models like RoBERTa, DistilBERT, etc...).
    cls_token = tokenizer.cls_token
    sep_token = tokenizer.sep_token
    # Candidate description is tokenized
    cand_tokens = tokenizer.tokenize(candidate_desc)

    # If a title is provided, tokenizes it, and then appropriately integrates it with the description tokens, adding a title_tag between them
    if candidate_title is not None:
        title_tokens = tokenizer.tokenize(candidate_title)
        if len(title_tokens) <= len(cand_tokens):
            cand_tokens = (
                title_tokens
                + [title_tag]
                + cand_tokens[
                    (
                        0
                        if title_tokens != cand_tokens[: len(title_tokens)]
                        else len(title_tokens)
                    ) :
                ]
            )  # Filter title from description
        else:
            cand_tokens = title_tokens + [title_tag] + cand_tokens

    cand_tokens = cand_tokens[: max_seq_length - 2]
    cand_tokens = [cls_token] + cand_tokens + [sep_token]

    # Token to ID Conversion and Padding
    input_ids = tokenizer.convert_tokens_to_ids(cand_tokens)
    padding = [0] * (max_seq_length - len(input_ids))
    input_ids += padding
    assert len(input_ids) == max_seq_length

    # Output : Returns the processed tokens and their corresponding IDs
    return {
        "tokens": cand_tokens,
        "ids": input_ids,
    }


def process_mention_data(
    samples,
    entity_dictionary,
    tokenizer,
    max_context_length,
    max_cand_length,
    silent,
    knn,
    dictionary_processed=False,
    mention_key="mention",
    context_key="context",
    label_key="label",
    multi_label_key=None,
    title_key="label_title",
    label_id_key="label_id",
    ent_start_token=ENT_START_TAG,
    ent_end_token=ENT_END_TAG,
    title_token=ENT_TITLE_TAG,
    debug=False,
    logger=None,
):
    """
    Description
    -----------
    Process the raw text mention data to a tokenized version and packages it into a form that can be directly used for training or inference in NLP models

    Parameters
    ----------
    samples : list of dicts
        Each dict represents a text sample with mention and context.
    entity_dictionary : dict
        A dictionary of entities with their descriptions, titles, etc.
    tokenizer : object
        A tokenizer object for converting text to tokens (like from BERT).
    max_context_length : int
        Maximum length of tokenized context.
    max_cand_length : int
        Maximum length for tokenized candidate entities.
    silent : bool
        If True, suppresses progress output.
    knn : int
        Number of nearest neighbors to consider in processing.
    dictionary_processed : bool, optional:
        Indicates if entity dictionary is already processed.
    mention_key, context_key, label_key, label_id_key : str
        Keys to access respective elements in samples.
    multi_label_key str, optional:
        Key for multiple labels, if applicable.
    title_key : str
        Key for accessing entity titles in the dictionary.
    ent_start_token, ent_end_token, title_token : str
        Special tokens used in tokenization.
    debug : bool, optional
        If True, processes a subset for debugging.
    logger : object, optional
        Logger object for logging information.

    -------
    Returns the processed samples (=mentions), updated entity dictionary (=tokenized entities), and tensor dataset (wrapping tensor into a dataset)
    """

    # Entity Dictionary Processing :
    # Tokenize entities from the entity dictionary if not already processed
    processed_samples = []
    dict_cui_to_idx = {}
    for idx, ent in enumerate(tqdm(entity_dictionary, desc="Tokenizing dictionary")):
        dict_cui_to_idx[ent["cui"]] = idx
        if not dictionary_processed:
            label_representation = get_candidate_representation(
                ent["description"], tokenizer, max_cand_length, ent["title"]
            )
            entity_dictionary[idx]["tokens"] = label_representation["tokens"]
            entity_dictionary[idx]["ids"] = label_representation["ids"]

    # Debugging test
    if debug:
        samples = samples[:200]

    iter_ = samples

    # Sample (= Mention data) Processing :
    # Processes each sample, tokenizing its context and resolving its labels to indices in the entity dictionary
    for idx, sample in enumerate(iter_):
        context_tokens = get_context_representation(
            sample,
            tokenizer,
            max_context_length,
            mention_key,
            context_key,
            ent_start_token,
            ent_end_token,
        )

        labels, record_labels, record_cuis = [sample], [], []
        # If the multi_label_key is provided (meaning samples can have multiple labels),
        # labels is set to the list of labels from the sample corresponding to this key.
        if multi_label_key is not None:
            labels = sample[multi_label_key]

        not_found_in_dict = False
        for l in labels:
            label = l[label_key]
            label_idx = l[label_id_key]
            if label_idx not in dict_cui_to_idx:
                not_found_in_dict = True
                break
            record_labels.append(dict_cui_to_idx[label_idx])
            record_cuis.append(label_idx)

        if not_found_in_dict:
            continue

        record = {
            "mention_id": sample.get("mention_id", idx),
            "mention_name": sample["mention"],  # The actual text of the mention
            "context": context_tokens,  # Contains the tokenized version of (mention + surrounding context + special tokens)
            "n_labels": len(
                record_labels
            ),  # Number of labels associated with the mention
            # An array of indices pointing to the entities in the entity dictionary that are considered correct labels for the mention.
            # This array is padded with -1 to reach a length of knn, which facilitates operations that consider a fixed number of potential labels for each mention
            "label_idxs": record_labels + [-1] * (knn - len(record_labels)),
            "label_cuis": record_cuis,  # CUIs associated with the correct labels for the mention
            "type": sample["type"],  # Represents the type of entity or mention
        }

        processed_samples.append(record)

    # Optionally logs processed samples for inspection
    if logger:
        logger.info("====Processed samples: ====")
        for sample in processed_samples[:5]:
            logger.info("Context tokens : " + " ".join(sample["context"]["tokens"]))
            logger.info(
                "Context ids : " + " ".join([str(v) for v in sample["context"]["ids"]])
            )
            for l in sample["label_idxs"]:
                if l == -1:
                    break
                logger.info(
                    f"Label {l} tokens : " + " ".join(entity_dictionary[l]["tokens"])
                )
                logger.info(
                    f"Label {l} ids : "
                    + " ".join([str(v) for v in entity_dictionary[l]["ids"]])
                )

    # Tensor Dataset Creation :
    # Converts processed samples into PyTorch tensors for model input.
    context_vecs = torch.tensor(
        select_field(processed_samples, "context", "ids"),
        dtype=torch.long,
    )
    label_idxs = torch.tensor(
        select_field(processed_samples, "label_idxs"),
        dtype=torch.long,
    )
    n_labels = torch.tensor(
        select_field(processed_samples, "n_labels"),
        dtype=torch.int,
    )
    mention_idx = torch.arange(len(n_labels), dtype=torch.long)

    tensor_data = TensorDataset(context_vecs, label_idxs, n_labels, mention_idx)

    return processed_samples, entity_dictionary, tensor_data


def process_mention_with_candidate(
    samples,
    entity_dictionary,
    tokenizer,
    max_context_length,
    max_cand_length,
    silent,
    dictionary_processed=False,
    mention_key="mention",
    context_key="context",
    label_key="label",
    multi_label_key=None,
    title_key="label_title",
    label_id_key="label_id",
    ent_start_token=ENT_START_TAG,
    ent_end_token=ENT_END_TAG,
    title_token=ENT_TITLE_TAG,
    debug=False,
    logger=None,
    drop_entities=[],
):
    # Entity Dictionary Processing :
    # Tokenize entities from the entity dictionary if not already processed
    processed_samples = []
    dict_cui_to_idx = {}
    for idx, ent in enumerate(tqdm(entity_dictionary, desc="Tokenizing dictionary")):
        dict_cui_to_idx[ent["cui"]] = idx
        if not dictionary_processed:
            label_representation = get_candidate_representation(
                ent["description"], tokenizer, max_cand_length, ent["title"]
            )
            entity_dictionary[idx]["tokens"] = label_representation["tokens"]
            entity_dictionary[idx]["ids"] = label_representation["ids"]

    if debug:
        samples = samples[:200]

    if silent:
        iter_ = samples
    else:
        iter_ = tqdm(samples)

    use_world = True

    id_to_idx = {}
    label_id_is_int = True

    drop_entities_set = set(drop_entities)

    for idx, sample in enumerate(iter_):
        if sample["label_id"] in drop_entities_set:
            continue

        context_tokens = get_context_representation(
            sample,
            tokenizer,
            max_context_length,
            mention_key,
            context_key,
            ent_start_token,
            ent_end_token,
        )

        label = sample[label_key]
        title = sample.get(title_key, None)

        label_tokens = get_candidate_representation(
            label,
            tokenizer,
            max_cand_length,
            title,
        )

        labels, record_labels, record_cuis = [sample], [], []
        # If the multi_label_key is provided (meaning samples can have multiple labels),
        # labels is set to the list of labels from the sample corresponding to this key.
        if multi_label_key is not None:
            labels = sample[multi_label_key]

        not_found_in_dict = False
        for l in labels:
            label = l[label_key]
            label_idx = l[label_id_key]
            if label_idx not in dict_cui_to_idx:
                not_found_in_dict = True
                break
            record_labels.append(dict_cui_to_idx[label_idx])
            record_cuis.append(label_idx)

        if not_found_in_dict:
            continue

        record = {
            "context": context_tokens,
            "label": label_tokens,
            "label_idx": [record_labels],
        }

        processed_samples.append(record)

    if logger:
        logger.info("====Processed samples: ====")
        for sample in processed_samples[:5]:
            logger.info("Context tokens : " + " ".join(sample["context"]["tokens"]))
            logger.info(
                "Context ids : " + " ".join([str(v) for v in sample["context"]["ids"]])
            )
            logger.info("Label tokens : " + " ".join(sample["label"]["tokens"]))
            logger.info(
                "Label ids : " + " ".join([str(v) for v in sample["label"]["ids"]])
            )
            logger.info(f"Label_id : {sample['label_idx'][0]}")

    context_vecs = torch.tensor(
        select_field(processed_samples, "context", "ids"),
        dtype=torch.long,
    )
    cand_vecs = torch.tensor(
        select_field(processed_samples, "label", "ids"),
        dtype=torch.long,
    )
    label_idx = torch.tensor(
        select_field(processed_samples, "label_idx"),
        dtype=torch.long,
    )
    data = {
        "context_vecs": context_vecs,  # Mention + context : Indices in the tokenizer vocabulary of "[CLS] c_left [START] mention [END] c_right [SEP]" for each sample in the test set
        "cand_vecs": cand_vecs,  # Ground truth entities : Indices in the tokenizer vocabulary of "[CLS] e_title [TITLE] e_desc [SEP]" for each sample in the test set
        "label_idx": label_idx,  # Indice of the correct entity
    }

    tensor_data = TensorDataset(context_vecs, cand_vecs, label_idx)
    return data, tensor_data


def compute_gold_clusters(mention_data):
    """
    Description
    -----------
    Group the mentions (in mention_data) into clusters based on their labels
    (Retrieves the position of each mention that shares the same label)

    Parameters
    ----------
    mention_data : list of dict
        List where each element is a dictionary representing a mention
    -------
    Returns a dict(list) where each key is a label index and each value is a list of mention indices that share that same label.
    Ex : clusters = {"label_1" : [2, 7, 12],
                    "label_2" : [3, 6, 10]}
    """
    clusters = {}
    for men_idx, mention in enumerate(mention_data):
        for i in range(mention["n_labels"]):
            label_idx = mention["label_idxs"][i]
            if label_idx not in clusters:
                clusters[label_idx] = []
            clusters[label_idx].append(men_idx)
    return clusters


def build_index(embeds, force_exact_search, probe_mult_factor=1):
    """
    Description
    -----------
    Constructs a search index using FAISS to efficiently perform similarity searches on a set of embeddings

    Parameters
    ----------
    embeds : numpy.ndarray
        Embeddings for which the index is to be built
    force_exact_search : bool
        Boolean flag indicating whether to use exact search instead of approximate search
    probe_mult_factor : int or float
        A multiplier factor used to determine the number of cells to probe in an approximate search
    -------
    Returns index (which is a FAISS index object). This index can be used for efficient similarity search among the embeddings.
    """
    # Convert Embeddings to NumPy Array (format required for further processing with faiss library)
    if type(embeds) is not np.ndarray:
        if torch.is_tensor(embeds):
            embeds = embeds.numpy()  # Converts pytorch tensor to a numpy array
        else:  # Could be a list, or any iterable
            embeds = np.array(embeds)

    # Initialize Index Variables
    nembeds = embeds.shape[0]  # number of embeddings
    d = embeds.shape[1]  # dimensionality of each embedding

    if (
        nembeds <= 10000 or force_exact_search
    ):  # if the number of embeddings is small, don't approximate
        index = faiss.IndexFlatIP(d)  # D,C
        index.add(embeds)
    else:
        # number of quantized clusters / cells that we want for the search
        nlist = int(math.floor(math.sqrt(nembeds)))
        # number of the quantized clusters / cells to probe (=to search)
        nprobe = int(math.floor(math.sqrt(nlist) * probe_mult_factor))
        quantizer = faiss.IndexFlatIP(d)
        index = faiss.IndexIVFFlat(
            quantizer, d, nlist, faiss.METRIC_INNER_PRODUCT
        )  # D,C
        index.train(embeds)  # D
        index.add(embeds)
        index.nprobe = nprobe
    return index


def embed_and_index(
    model,
    token_id_vecs,
    encoder_type,
    batch_size=768,
    only_embed=False,
    corpus=None,
    force_exact_search=False,
    probe_mult_factor=1,
    # world_size=1,
):
    """
    Description
    -----------
    Designed for embedding the token ID vectors AND INDEXING them for efficient search (tldr : embedding + indexing)

    Parameters
    ----------
    model :
        Object representing a ML model
    token_id_vecs : tensor
        Tensors representing tokenized text data
    encoder_type : str
        Determines the type of encoder to be used from the model. It must be either "context" or "candidate".
    batch_size : int
        The number of token vectors to process in each batch.
    only_embed : bool
        If set to True, the function returns only the embeddings, skipping the indexing part.
    corpus : list of dict
        Collection of entities, which is used to build type-specific search indexes if provided.
    force_exact_search : bool
        Determine whether to use exact search methods or approximate methods while building the search index.
    probe_mult_factor : int
        A multiplier factor used in index building for probing in case of approximate search
    -------
    Returns :
        The embeddings of the token ID vectors with dim = (len(token_id_vecs), embedding_dimension)
        If encoder_type is "context", embeds contains the embeddings of : [CLS] c_left [START] mention [END] c_right [SEP]
        If encoder_type is "candidate", embeds contains the embeddings of [CLS] e_title [TITLE] e_desc [SEP]

        And optionally : Builds and returns a search index based on the embeddings.
    """

    with torch.no_grad():
        # Selects the appropriate encoder
        if encoder_type == "context":  # mention
            print("encode context")
            encoder = model.encode_context
        elif encoder_type == "candidate":  # entity
            print("encode candidate")
            encoder = model.encode_candidate
        else:
            raise ValueError("Invalid encoder_type: expected context or candidate")

        # Compute embeddings
        embeds = None
        sampler = SequentialSampler(token_id_vecs)
        # sampler = (
        #     DistributedSampler(token_id_vecs, num_replicas=world_size, rank=get_rank())
        #     if world_size > 1
        #     else SequentialSampler(token_id_vecs)
        # )
        dataloader = DataLoader(token_id_vecs, sampler=sampler, batch_size=batch_size)
        iter_ = tqdm(dataloader, desc="Embedding in batches")
        for step, batch in enumerate(iter_):
            batch_embeds = encoder(
                batch.cuda()
            )  # After encoding, it's being sent back to cpu
            embeds = (
                batch_embeds
                if embeds is None
                else np.concatenate((embeds, batch_embeds), axis=0)
            )

        # # If using multiple GPUs, gather all embeddings on each process
        # if world_size > 1:
        #     # Convert numpy array to tensor for gathering
        #     embeds_tensor = torch.tensor(embeds).cuda()
        #     # Prepare a list to gather all tensor embeddings
        #     gather_list = [torch.zeros_like(embeds_tensor) for _ in range(world_size)]
        #     all_gather(gather_list, embeds_tensor)
        #     # Concatenate all gathered tensors and convert back to numpy
        #     embeds = torch.cat(gather_list, dim=0).cpu().numpy()
        #     # Trim the extra samples from DistributedSampler padding
        #     embeds = embeds[: token_id_vecs.size(0)]

        if isinstance(embeds, torch.Tensor):
            embeds = embeds.numpy()

        if only_embed:
            return embeds

        if corpus is None:
            # When "use_types" is False
            index = build_index(
                embeds, force_exact_search, probe_mult_factor=probe_mult_factor
            )
            return embeds, index

        # Build type-specific search indexes
        search_indexes = (
            {}
        )  # Dictionary that will store search indexes (!= indices)for each unique entity type found in the corpus
        corpus_idxs = (
            {}
        )  # Dictionary to store indices of the corpus elements, grouped by their entity type.
        for i, e in enumerate(corpus):
            ent_type = e["type"]
            if ent_type not in corpus_idxs:
                corpus_idxs[ent_type] = []
            corpus_idxs[ent_type].append(i)
        for ent_type in corpus_idxs:
            search_indexes[ent_type] = build_index(
                embeds[corpus_idxs[ent_type]],
                force_exact_search,
                probe_mult_factor=probe_mult_factor,
            )
            corpus_idxs[ent_type] = np.array(corpus_idxs[ent_type])

        return embeds, search_indexes, corpus_idxs


def get_index_from_embeds(
    embeds, corpus_idxs=None, force_exact_search=False, probe_mult_factor=1
):
    """
    Description
    -----------
    Designed to build search indexes from embeddings.
    Similar to "build_index" function but it can handle the creation of multiple indexes if corpus_idxs is provided (set of embeddings).

    Parameters
    ----------
    embeds : numpy.ndarray
        Embeddings for which the index is to be built
    corpus_idxs : dict
        Dictionary containing entity types (each entity type have the indices from the corpus corresponding to that entity)
    force_exact_search : bool
        Determine whether to use exact search methods or approximate methods while building the search index.
    probe_mult_factor : int
        A multiplier factor used in index building for probing in case of approximate search
    -------
    Returns a single index if corpus_idxs is None, or a dictionary of indexes (one for each entity type) if corpus_idxs is provided.
    """
    if corpus_idxs is None:
        index = build_index(
            embeds, force_exact_search, probe_mult_factor=probe_mult_factor
        )
        return index
    search_indexes = {}
    for ent_type in corpus_idxs:
        search_indexes[ent_type] = build_index(
            embeds[corpus_idxs[ent_type]],
            force_exact_search,
            probe_mult_factor=probe_mult_factor,
        )
    return search_indexes


def get_idxs_by_type(corpus):
    """
    Description
    -----------
    Regroup the entities INDICES based on entity types
    It simply organizes the data.

    Parameters
    ----------
    corpus : list of dict
        Collection of entities.
    -------
    Returns the corpus_idxs dictionary, which maps each entity type to an array of INDICES representing the positions of entities of that type in the original corpus
    Ex : entity_dict = {"entity_1" : [2, 7, 12], #indices
                    "entity_2" : [3, 6, 10]}
    """
    corpus_idxs = (
        {}
    )  # Dictionary used to store indices of the corpus elements, grouped by their entity type.
    for i, e in enumerate(corpus):
        ent_type = e["type"]
        if ent_type not in corpus_idxs:
            corpus_idxs[ent_type] = []
        corpus_idxs[ent_type].append(i)
    for ent_type in corpus_idxs:
        corpus_idxs[ent_type] = np.array(corpus_idxs[ent_type])
    return corpus_idxs



import os
import sys

import numpy as np
from tqdm import tqdm
import pickle
import argparse
from typing import Optional

import torch
import torch.nn as nn
from torch.utils.data import DataLoader, Dataset
import pytorch_lightning as pl
from transformers import TrainingArguments
from transformers import BartTokenizer
import torch 
import torch.nn as nn
from transformers import TrainingArguments
from transformers import Seq2SeqTrainingArguments
from bioel.models.biogenel.trainer import modifiedSeq2SeqTrainer
from bioel.models.biogenel.trie import Trie
from bioel.models.biogenel.utils import reform_input
from bioel.models.biogenel.LightningDataModule import BioGenELDataModule
from bioel.models.biogenel.fairseq_beam import (SequenceGenerator,PrefixConstrainedBeamSearch,PrefixConstrainedBeamSearchWithSampling)
import copy
import json
import ujson

from bioel.ontology import BiomedicalOntology

class BioGenElLightningModule(pl.LightningModule):
    def __init__(self, config, datamodule: Optional[BioGenELDataModule] = None):
        super().__init__()
        self.config = config
        if datamodule:
            self.datamodule = datamodule
        self.trie = None  # Initialize trie
        config.max_steps = config.max_steps // config.gradient_accumulate
        config.save_steps = config.max_steps

        self.training_args = Seq2SeqTrainingArguments(
                        output_dir=config.model_save_path,          # output directory
                        num_train_epochs=config.num_train_epochs,              # total number of training epochs
                        per_device_train_batch_size=config.per_device_train_batch_size,  # batch size per device during training
                        per_device_eval_batch_size=config.per_device_eval_batch_size,   # batch size for evaluation
                        warmup_steps=config.warmup_steps,                # number of warmup steps for learning rate scheduler
                        weight_decay=config.weight_decay,               # strength of weight decay
                        logging_dir=config.logging_path,            # directory for storing logs
                        logging_steps=config.logging_steps,
                        save_steps=config.save_steps,
                        evaluation_strategy=config.evaluation_strategy,
                        learning_rate=config.init_lr,
                        label_smoothing_factor=config.label_smoothing_factor,
                        max_grad_norm=config.max_grad_norm,
                        max_steps=config.max_steps,
                        lr_scheduler_type=config.lr_scheduler_type,
                        seed=config.seed,
                        gradient_accumulation_steps=config.gradient_accumulate,
                        generation_config = None,
                        save_safetensors=False,
                        )

        # Initialize model and tokenizer based on config
        if config.evaluation:
            if config.t5:
                from bioel.models.biogenel.models import T5EntityPromptModel
                from transformers import T5Tokenizer, T5Config
                #If one day we can get the code for prepare_trainer_dataset_t5
                #from bioel.models.biogenel.LightningDataModule import prepare_trainer_dataset_t5 as prepare_trainer_dataset

                t5conf = T5Config.from_pretrained("./t5-large")
                t5conf.dropout_rate = config.dropout

                self.tokenizer = T5Tokenizer.from_pretrained("./t5-large")

                self.model = T5EntityPromptModel.from_pretrained(
                    config.model_load_path,
                    config=t5conf,
                    n_tokens=(config.prompt_tokens_enc, config.prompt_tokens_dec),
                    load_prompt=True,
                    soft_prompt_path=config.model_load_path,
                )
            else:
                from bioel.models.biogenel.models import BartEntityPromptModel
                from transformers import BartTokenizer, BartConfig
                #from bioel.models.biogenel.LightningDataModule import prepare_trainer_dataset_fine as prepare_trainer_dataset

                self.tokenizer = BartTokenizer.from_pretrained(config.model_token_path)

                bartconf = BartConfig.from_pretrained(config.model_load_path)
                bartconf.max_position_embeddings = config.max_position_embeddings
                bartconf.attention_dropout = config.attention_dropout
                bartconf.dropout = config.dropout
                bartconf.max_length = config.max_length

                self.model = BartEntityPromptModel.from_pretrained(
                    config.model_load_path,
                    config=bartconf,
                    n_tokens=(config.prompt_tokens_enc, config.prompt_tokens_dec),
                    load_prompt=True,
                    soft_prompt_path=config.model_load_path,
                )
        else:
            if config.t5:
                # Initialize T5 model and tokenizer
                from bioel.models.biogenel.models import T5EntityPromptModel
                from transformers import T5Tokenizer, T5Config
                #If one day we can get the code for prepare_trainer_dataset_t5
                #from bioel.models.biogenel.LightningDataModule import prepare_trainer_dataset_t5 as prepare_trainer_dataset

                t5conf = T5Config.from_pretrained('./t5-large')
                t5conf.dropout_rate = config.dropout
                self.tokenizer = T5Tokenizer.from_pretrained('./t5-large')

                self.model = T5EntityPromptModel.from_pretrained(config.model_load_path, 
                                                            config = t5conf,
                                                            finetune = config.finetune, 
                                                            n_tokens = (config.prompt_tokens_enc, config.prompt_tokens_dec),
                                                            load_prompt = config.load_prompt,
                                                            soft_prompt_path = config.model_load_path,
                                                            initialize_from_vocab = config.init_from_vocab,
                                                            )
            else:
                # Initialize Bart model and tokenizer
                from bioel.models.biogenel.models import BartEntityPromptModel
                from transformers import BartTokenizer, BartConfig

                bartconf = BartConfig.from_pretrained(config.model_load_path)
                bartconf.max_position_embeddings = config.max_position_embeddings
                bartconf.attention_dropout = config.attention_dropout
                bartconf.dropout = config.dropout

                self.tokenizer = BartTokenizer.from_pretrained(config.model_token_path, 
                                                        max_length=1024,
                                                        )

                self.model = BartEntityPromptModel.from_pretrained(config.model_load_path, 
                                                            config = bartconf,
                                                            finetune = config.finetune, 
                                                            n_tokens = (config.prompt_tokens_enc, config.prompt_tokens_dec),
                                                            load_prompt = config.load_prompt,
                                                            soft_prompt_path = config.model_load_path,
                                                            no_finetune_decoder = config.no_finetune_decoder,
                                                            )
            
        self.model = self.model.cuda().to(self.model.device) 

            
    def trainer_choice(self):
        if self.config.preprocess_data:
            self.datamodule.prepare_data()
        self.datamodule.setup()
        train_dataset = self.datamodule.train_dataloader().dataset
        if self.config.unlikelihood_loss:
            print("loading trie......")
            with open(self.config.trie_path, "rb") as f:
                self.trie = Trie.load_from_dict(pickle.load(f))
            print("trie loaded.......")
            

            self.trainer = modifiedSeq2SeqTrainer(
                        model=self.model,                         # the instantiated  Transformers model to be trained
                        args=self.training_args,                  # training arguments, defined above
                        train_dataset=train_dataset, 
                        fairseq_loss=self.config.fairseq_loss,  
                        enc_num = self.config.prompt_tokens_enc,
                        dec_num = self.config.prompt_tokens_dec,
                        prefix_allowed_tokens_fn = lambda batch_id, sent: self.trie.get(sent.tolist()),
                        rdrop = self.config.rdrop,
                    )
        else:
            self.trainer = modifiedSeq2SeqTrainer(
                                    model=self.model,                         # the instantiated  Transformers model to be trained
                                    args=self.training_args,                  # training arguments, defined above
                                    train_dataset=train_dataset, 
                                    fairseq_loss=self.config.fairseq_loss,  
                                    enc_num = self.config.prompt_tokens_enc,
                                    dec_num = self.config.prompt_tokens_dec,
                                    rdrop = self.config.rdrop,
                                )

    def train(self):
        self.trainer_choice()
        self.trainer.train()

    def evaluate(self):
        self.evaluate_choice()
        self.on_evaluate()
        self.apply_fairseq_beam_search()
    
    def init_datamodule(self, datamodule = BioGenELDataModule):
        self.datamodule = datamodule

    def evaluate_choice(self):
        if self.config.preprocess_data:
            self.datamodule.prepare_data()
        self.datamodule.setup()
        dataframe = self.datamodule.deduplicated
        
        if self.config.testset:
            print('eval on test set')
            self.eval_dataset = self.datamodule.test_dataloader().dataset
            self.eval_docs = dataframe[dataframe['split'] == 'test']
        else:
            print('eval on develop set')
            self.eval_dataset = self.datamodule.dev_dataloader().dataset
            self.eval_docs = dataframe[dataframe['split'] == 'validation']
    
    def on_evaluate(self):
        print("loading cui2str dictionary....")
        dict_path = self.config.dict_path
        print("dictpath", dict_path)
        if "json" in dict_path:
            with open(dict_path, "r") as f:
                cui2str = ujson.load(f)
        else:
            with open(dict_path, "rb") as f:
                cui2str = pickle.load(f)

        str2cui = {}
        for cui in cui2str:
            if isinstance(cui2str[cui], list):
                for name in cui2str[cui]:
                    if name in str2cui:
                        str2cui[name].append(cui)
                    else:
                        str2cui[name] = [cui]
            else:
                name = cui2str[cui]
                if name in str2cui:
                    str2cui[name].append(cui)
                    print("duplicated vocabulary")
                else:
                    str2cui[name] = [cui]
        print("dictionary loaded......")

        if self.config.rerank:
            print("loading retrieved names......")
            with open(self.config.retrieved_path, "r") as f:
                retrieved_names = [line.split("\t")[0].split(" ") for line in f.readlines()]
            print("retrieved names loaded.")
            for i, l in tqdm(enumerate(retrieved_names)):
                for cui in list(l):
                    if cui in cui2str:
                        continue
                    else:
                        retrieved_names[i].remove(cui)

            print("loading tokenized names......")
            with open(self.config.dataset_path + "/tokenized.json", "r") as f:
                tokenized_names = ujson.load(f)
            print("tokenized names loaded.")
            self.retrieved_names = retrieved_names
            self.tokenized_names = tokenized_names

        if self.config.gold_sty:
            print("loading tokenized names......")
            with open(self.config.dataset_path + "/tokenized.json", "r") as f:
                tokenized_names = ujson.load(f)
            print("tokenized names loaded.")

            print("loading sty to cui dict.....")
            with open(self.config.dataset_path + "/sty2cui.json", "r") as f:
                sty2cuis = ujson.load(f)
            with open(self.config.dataset_path + "/sty.json", "r") as f:
                cuis2sty = ujson.load(f)
            print("sty to cui dict loaded.")

            trie_dict = {}
            for sty in sty2cuis:
                names = []
                for cui in tqdm(sty2cuis[sty]):
                    names += tokenized_names[cui]
                trie_dict[sty] = Trie(names)
            
            self.tokenized_names = tokenized_names
            self.cuis2sty = cuis2sty
            self.sty2cuis = sty2cuis
            self.trie_dict = trie_dict

        print("loading trie......")
        with open(self.config.trie_path, "rb") as f:
            trie = Trie.load_from_dict(pickle.load(f))
        print("trie loaded.......")

        print("loading label cuis......")
        with open(self.config.dataset_path + "/testlabel.txt", "r") as f:
            cui_labels = [
                set(cui.strip("\n").replace("+", "|").split("|")) for cui in f.readlines()
            ]
        print("label cuis loaded")

        self.str2cui = str2cui
        self.trie = trie
        self.cui_labels = cui_labels 


    def apply_fairseq_beam_search(self):
        if self.config.beam_threshold == 0:
            print('without using beam threshold')
            beam_strategy = PrefixConstrainedBeamSearch(
                tgt_dict=None, 
                prefix_allowed_tokens_fn=lambda batch_id, sent: self.trie.get(sent.tolist())
            )
        else:
            beam_strategy = PrefixConstrainedBeamSearchWithSampling(
                tgt_dict=None, 
                prefix_allowed_tokens_fn=lambda batch_id, sent: self.trie.get(sent.tolist()),
                logit_thresholding=self.config.beam_threshold,
            )
        
        fairseq_generator = SequenceGenerator(
            models = self.model,
            tgt_dict = None,
            beam_size=self.config.num_beams,
            max_len_a=0,
            max_len_b=self.config.max_length,
            min_len=self.config.min_length,
            eos=self.model.config.eos_token_id,
            search_strategy=beam_strategy,

            ##### all hyperparams below are set to default
            normalize_scores=True,
            len_penalty=self.config.length_penalty,
            unk_penalty=0.0,
            temperature=1.0,
            match_source_len=False,
            no_repeat_ngram_size=0,
            symbols_to_strip_from_output=None,
            lm_model=None,
            lm_weight=1.0,
        )

        results = list()
        cui_results = list()
        results_score = list()
        text = list()
        candidates_metadata = list()
        output = list()

        input_ids = []
        decoder_input_ids = []
        attention_mask = []
        count_top1 = 0
        count_top5 = 0
        for i in tqdm(range(0, len(self.eval_dataset))):
            
            if self.config.rerank:
                self.trie = Trie(sum([self.tokenized_names[cui] for cui in self.retrieved_names[i]], []))
                fairseq_generator.search = (PrefixConstrainedBeamSearch(
                                                                        tgt_dict=None, 
                                                                        prefix_allowed_tokens_fn=lambda batch_id, sent: self.trie.get(sent.tolist())
                                                                        ))
            if self.config.gold_sty:
                self.trie = self.trie_dict[self.cuis2sty[self.cui_labels[i]]]
                fairseq_generator.search = (PrefixConstrainedBeamSearch(
                                                                        tgt_dict=None, 
                                                                        prefix_allowed_tokens_fn=lambda batch_id, sent: self.trie.get(sent.tolist())
                                                                        ))
                
            input_ids.append(self.eval_dataset[i]['input_ids'])
            attention_mask.append(self.eval_dataset[i]['attention_mask'])
            decoder_input_ids.append(self.eval_dataset[i]['decoder_input_ids_test'])

            if i%self.config.per_device_eval_batch_size == 0:

                input_ids, attention_mask = reform_input(torch.stack(input_ids), attention_mask=torch.stack(attention_mask), ending_token=self.model.config.eos_token_id)
                sample = {'net_input':{'input_ids':input_ids, 'attention_mask':attention_mask}}
                
                result_tokens, posi_scores = fairseq_generator.forward(
                    sample=sample,
                    prefix_mention_is = self.config.prefix_mention_is,
                    prefix_tokens=decoder_input_ids[0].unsqueeze(0).cuda() if self.config.prefix_mention_is else None,
                )

                for ba, beam_sent in enumerate(result_tokens):
                    result = []
                    cui_result = []
                    candidates = []
                    for be, sent in enumerate(beam_sent):
                        if self.config.prefix_mention_is:
                            result.append(self.tokenizer.decode(sent[len(decoder_input_ids[0]):], skip_special_tokens=True))
                        else:
                            result.append(self.tokenizer.decode(sent, skip_special_tokens=True))
                    
                    for idx, r in enumerate(result):
                        if r.strip(" ") in self.str2cui:
                            if idx == 0:
                                text.append(r.strip(" "))
                            cui_result.append(self.str2cui[r.strip(" ")])
                            candidates.append(
                                {"text": r.strip(" "), "db_id": self.str2cui[r.strip(" ")]}
                            )
                        else:
                            print(f"Candidate not found in str2cui: {r}")

                            if (
                                "MESH" in r
                                or "mesh" in r
                                or "omim" in r
                                or "OMIM" in r
                                or "umls" in r
                                or "UMLS" in r
                                or "NCBIGene" in r
                                or "NCBI" in r
                            ):
                                if idx == 0:
                                    text.append(r)
                                cui_result.append(r)
                                candidates.append({"text": r.strip(" "), "db_id": r})
                            else:
                                continue

                    if len(cui_result) == 0:
                        continue

                    cui_results.append(cui_result)
                    results.append(result)
                    results_score.append(posi_scores)
                    candidates_metadata.append(candidates)
                    if 'deabbreviated_text' in self.eval_docs:
                        output.append(
                            {
                                "document_id": self.eval_docs['document_id'].iloc[i],
                                "offsets": self.eval_docs['offsets'].iloc[i],
                                "text": self.eval_docs['text'].iloc[i],
                                "type": self.eval_docs['type'].iloc[i],
                                "db_ids": list(self.cui_labels[i]),
                                "split": self.eval_docs['split'].iloc[i],
                                "deabbreviated_text":self.eval_docs['deabbreviated_text'].iloc[i],
                                "mention_id": self.eval_docs['mention_id'].iloc[i] + ".abbr_resolved",
                                "candidates": cui_result,
                                "candidates_metadata": candidates,
                            })
                    else:
                        output.append(
                        {
                            "document_id": self.eval_docs['document_id'].iloc[i],
                            "offsets": self.eval_docs['offsets'].iloc[i],
                            "text": self.eval_docs['text'].iloc[i],
                            "type": self.eval_docs['type'].iloc[i],
                            "db_ids": list(self.cui_labels[i]),
                            "split": self.eval_docs['split'].iloc[i],
                            "mention_id": self.eval_docs['mention_id'].iloc[i],
                            "candidates": cui_result,
                            "candidates_metadata": candidates,
                        })
                    # print(self.tokenizer.decode(input_ids[0], skip_special_tokens=True))
                    # print(posi_scores)
                    # print(result)
                    # print(cui_result)
                    # print(self.cui_labels[i])
                    # print(self.cui2str[self.cui_labels[i]])
                    # input()

                    if self.cui_labels[i].intersection(set(cui_result[0])):
                        count_top1 += 1
                        count_top5 += 1
                    elif self.cui_labels[i].intersection(set(sum(cui_result,[]))):
                        count_top5 += 1

                if i % 50 == 49:
                    print('=============Top1 Precision:\t',count_top1/(i+1))
                    print('=============Top5 Precision:\t',count_top5/(i+1))

                input_ids = []
                decoder_input_ids = []
                attention_mask = []
        
        print("last")
        print('=============Top1 Precision:\t',count_top1/(i+1))
        print('=============Top5 Precision:\t',count_top5/(i+1))

        with open('./logs.txt', 'a+') as f:
            f.write(str(self.config.seed) + '======\n')
            f.write(self.config.model_load_path + '\n')
            f.write('Top1 Precision:\t'+str(count_top1/(i+1))+'\n')
            f.write('Top5 Precision:\t'+str(count_top5/(i+1))+'\n\n')

        with open(os.path.join(self.config.model_load_path, 'output.json'), 'w') as f:
            f.write(ujson.dumps(output, indent=2))
        
        if self.config.testset:

            with open(os.path.join(self.config.model_load_path, 'results_test.pkl'), 'wb') as f:
                pickle.dump([results, results_score], f)

        else:

            with open(os.path.join(self.config.model_load_path, 'results_dev.pkl'), 'wb') as f:
                pickle.dump([results, results_score], f)
    
    def train_dataloader(self):
        return self.datamodule.train_dataloader()

    def val_dataloader(self):
        return self.datamodule.dev_dataloader()

    def test_dataloader(self):
        return self.datamodule.test_dataloader()
    

if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='training configuration')

    parser.add_argument("dataset_path",type=str,
                        help="path of the medmentions dataset")
    parser.add_argument("-model_save_path",type=str,default='./model_saved',
                        help="path of the pretrained model")
    parser.add_argument("-trie_path",type=str,default='./trie.pkl',
                        help="path of the Trie")
    parser.add_argument("-dict_path",type=str,default='./trie.pkl',
                        help="path of the cui2str dictionary")
    parser.add_argument("-retrieved_path",type=str,default='./trie.pkl',
                        help="path of the cui2str dictionary")
    parser.add_argument("-model_load_path",type=str,default='facebook/bart-large',
                        help="path of the pretrained model")
    parser.add_argument("-model_token_path",type=str,default='facebook/bart-large',
                        help="path of the pretrained model")
    parser.add_argument("-logging_path",type=str, default='./logs',
                        help="path of saved logs")
    parser.add_argument('-logging_steps', type=int, default=500,
                        help='save logs per logging step')                  
    parser.add_argument('-save_steps', type=int, default = 20000,
                        help='save checkpoints per save steps')
    parser.add_argument('-num_train_epochs', type=int, default = 8,
                        help="number of training epochs")
    parser.add_argument('-per_device_train_batch_size', type=int, default = 4,
                        help='training batch size')
    parser.add_argument('-per_device_eval_batch_size', type=int, default = 5,
                        help='evaluation batch size')
    parser.add_argument('-warmup_steps', type=int, default = 500,
                        help='warmup steps')
    parser.add_argument('-finetune', action='store_true',
                        help='if finetune the bart params')
    parser.add_argument('-t5', action='store_true',
                        help='if use t5 pretrained model')
    parser.add_argument('-fairseq_loss', action='store_true',
                        help='if use label smoothed loss in fairseq')
    parser.add_argument('-evaluation', action='store_true',
                        help='whether to train or evaluate')
    parser.add_argument('-testset', action='store_true',
                        help='whether evaluate with testset or devset')
    parser.add_argument('-load_prompt', action='store_true',
                        help='whether to load prompt')
    parser.add_argument('-weight_decay', type=float, default = 0.01,
                        help='weigth decay of optimizer')
    parser.add_argument('-length_penalty', type=float, default = 1,
                        help='length penaltyof beam search')
    parser.add_argument('-beam_threshold', type=float, default = 0,
                        help='logit threshold of beam search')
    parser.add_argument('-unlikelihood_loss', action='store_true',
                        help='whether using unlikelihood loss')
    parser.add_argument('-init_lr', type=float, default = 5e-5,
                        help='initial learning rate of AdamW')
    parser.add_argument('-evaluation_strategy', type=str, default='no',
                        help='evaluation strategy')
    parser.add_argument('-prompt_tokens_enc', type=int, default = 0,
                        help='a tuple containing number of soft prompt tokens in encoder and decoder respectively')
    parser.add_argument('-prompt_tokens_dec', type=int, default = 0,
                        help='a tuple containing number of soft prompt tokens in encoder and decoder respectively')
    parser.add_argument('-seed', type=int, default = 42,
                        help='the seed of huggingface seq2seq training, 42 is also the default of huggingface train default')
    parser.add_argument('-label_smoothing_factor', type=float, default = 0.1,
                        help='label smoothig factor')
    parser.add_argument('-unlikelihood_weight', type=float, default = 0.1,
                        help='label smoothig factor')
    parser.add_argument('-max_grad_norm', type=float, default = 0.1,
                        help='gradient clipping value')
    parser.add_argument('-max_steps', type=int, default = 200000,
                        help='max training steps override num_train_epoch')
    parser.add_argument('-gradient_accumulate', type=int, default = 1,
                        help='max training steps override num_train_epoch')
    parser.add_argument('-lr_scheduler_type', type=str, default = 'polynomial',
                        help='the learning rate schedule type')
    parser.add_argument('-attention_dropout', type=float, default = 0.1,
                        help='the attention dropout')
    parser.add_argument('-dropout', type=float, default = 0.1,
                        help='dropout')
    parser.add_argument('-max_position_embeddings', type=int, default = 1024,
                        help="the max length for position embedding")
    parser.add_argument('-num_beams', type=int, default = 5,
                        help='the attention dropout')
    parser.add_argument('-max_length', type=int, default = 1024,
                        help='the attention dropout')
    parser.add_argument('-min_length', type=int, default = 1,
                        help='the attention dropout')
    parser.add_argument('-sample_train', action='store_true',
                        help='if to use training target sampled by tfidf similarity')
    parser.add_argument('-prefix_prompt', action='store_true',
                        help='wheather use prefix prompt tokens')
    parser.add_argument('-rerank', action='store_true',
                        help='wheather to rerank the retrieved names')
    parser.add_argument('-init_from_vocab', action='store_true',
                        help='wheather initialize prompt from the mean of token embeddings')
    parser.add_argument('-no_finetune_decoder', action='store_true',
                        help='whether only finetune encoder')
    parser.add_argument('-syn_pretrain', action='store_true',
                        help='whether only finetune encoder')
    parser.add_argument('-gold_sty', action='store_true',
                        help='whether only finetune encoder')
    parser.add_argument('-prefix_mention_is', action='store_true',
                        help='whether only finetune encoder')
    parser.add_argument('-rdrop', type=float, default=0.0)
    parser.add_argument('-preprocess_data', type=bool, default = False,
                        help='preprocess the data or not')



import os
import ujson
import warnings
from collections import defaultdict
import pandas as pd

# from bigbio.dataloader import BigBioConfigHelpers
import json
import joblib
import numpy as np
import datetime

from tqdm.auto import tqdm

tqdm.pandas()

from datasets import load_dataset
from bioel.utils.bigbio_utils import (
    dataset_to_documents,
    dataset_to_df,
    resolve_abbreviation,
    load_bigbio_dataset,
)
from bioel.utils.dataset_consts import (
    CUIS_TO_REMAP,
    CUIS_TO_EXCLUDE,
    VALIDATION_DOCUMENT_IDS,
)
from sklearn.feature_extraction.text import TfidfVectorizer

from bioel.ontology import BiomedicalOntology
from typing import List


def cuis_to_aliases(ontology: BiomedicalOntology, save_dir: str, dataset_name: str):
    """
    Creates the .txt file that maps the Cuis to the aliases and canonical name for each entities
    from the provided ontology.

    Parameters
    ----------
    ontology_dir: str, required.
        The path where the ontology is stored.
    save_dir : str, required.
        The path where the mapping file "{dataset_name}_aliases.txt" will be stored
    dataset_name : str, required.
        The name of the dataset for which the mapping will be created.
    """
    # Create the directory if it doesn't exist
    os.makedirs(save_dir, exist_ok=True)
    import re

    file_path = os.path.join(save_dir, f"{dataset_name}_aliases.txt")
    with open(file_path, "w", encoding="utf-8") as file:
        for cui, entity in ontology.entities.items():
            file.write(f"{cui}||{entity.name}\n")
            if len(entity.aliases) != 0:
                if isinstance(entity.aliases, list):
                    for word in entity.aliases:
                        if word is not None:
                            # print(word)
                            word = word.strip()
                            file.write(f"{cui}||{word}\n")
                else:
                    # words = entity.aliases.split('; , ')
                    words = re.split("[;|]", entity.aliases)
                    for word in words:
                        word = word.strip()
                        file.write(f"{cui}||{word}\n")
            if entity.equivalant_cuis:
                for eqcui in entity.equivalant_cuis:
                    if eqcui != entity.cui:
                        file.write(f"{eqcui}||{entity.name}\n")
                        if entity.aliases:
                            words = re.split("[;|]", entity.aliases)
                            for word in words:
                                word = word.strip()
                                file.write(f"{eqcui}||{word}\n")

    return file_path


if __name__ == "__main__":
    # data_preprocess(
    #     dataset_name = "bc5cdr",
    #     save_dir = "data2/",
    #     ontology_dir = "Y:/mitchell/entity-linking/2017AA/META/",
    #     ontology_name = "mesh",
    #     resolve_abbrevs=False)
    cuis_to_aliases(
        ontology=BiomedicalOntology.load_medic(
            filepath="/mitchell/entity-linking/kbs/medic.tsv", name="medic"
        ),
        save_dir="/home2/cye73/data_test2/sapbert/ncbi_disease/",
        dataset_name="ncbi_disease",
    )
    # cuis_to_aliases(
    #     ontology=BiomedicalOntology.load_entrez(
    #         filepath="/mitchell/entity-linking/el-robustness-comparison/data/gene_info.tsv",
    #         dataset="gnormplus",
    #     ),
    #     save_dir="/home2/cye73/data_test2/sapbert/ncbi_disease/",
    #     dataset_name="gnormplus",
    # )
    # cuis_to_aliases(
    #     ontology=BiomedicalOntology.load_entrez(
    #         filepath="/mitchell/entity-linking/el-robustness-comparison/data/gene_info.tsv",
    #         dataset="nlm_gene",
    #     ),
    #     save_dir="/home2/cye73/data_test2/sapbert/ncbi_disease/",
    #     dataset_name="nlm_gene",
    # )
    # cuis_to_aliases(
    #     ontology=BiomedicalOntology.load_mesh(
    #         filepath="/mitchell/entity-linking/2017AA/META/", name="mesh"
    #     ),
    #     save_dir="/home2/cye73/data_test2/sapbert/ncbi_disease/",
    #     dataset_name="bc5cdr",
    # )
    # cuis_to_aliases(
    #     ontology=BiomedicalOntology.load_mesh(
    #         filepath="/mitchell/entity-linking/2017AA/META/", name="mesh"
    #     ),
    #     save_dir="/home2/cye73/data_test2/sapbert/ncbi_disease/",
    #     dataset_name="nlm_chem",
    # )

    # cuis_to_aliases(
    #     ontology=BiomedicalOntology.load_umls(
    #         filepath="/mitchell/entity-linking/2017AA/META/", name="umls"
    #     ),
    #     save_dir="/home2/cye73/data_test2/sapbert/ncbi_disease/",
    #     dataset_name="medmentions_full",
    # )

    # cuis_to_aliases(
    #     ontology=BiomedicalOntology.load_umls(
    #         filepath="/mitchell/entity-linking/2017AA/META/", name="umls"
    #     ),
    #     save_dir="/home2/cye73/data_test2/sapbert/ncbi_disease/",
    #     dataset_name="medmentions_st21pv",
    # )

    # onto = BiomedicalOntology.load_entrez(
    #     filepath="/mitchell/entity-linking/el-robustness-comparison/data/gene_info.tsv",
    #     dataset="gnormplus",
    # )



import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt


def get_overlap(df, dataset_name=None, plot=False):
    """
    Overlap between train mentions and test mentions.

    Return both the % of entities that overlap and the % of mentions with entities that appear in the train set (since some entities appear more than once)
    """
    # Convert db_ids and type to sets
    df["db_ids_set"] = df["db_ids"].apply(set)
    df["type_set"] = df["type"].apply(set)

    # Get unique entities in data splits
    train_df = df.query('split == "train"')
    test_df = df.query('split == "test"')
    train_valid_df = df.query('split != "test"')

    # Flatten db_ids lists and convert to sets for uniqueness
    unique_train = set().union(*train_df.db_ids_set)
    unique_test = set().union(*test_df.db_ids_set)
    unique_train_valid = set().union(*train_valid_df.db_ids_set)
    unique_ents = set().union(*df.db_ids_set)

    # Flatten type lists for calculations
    all_types = df.explode("type")["type"]
    unique_types = set(all_types.unique())

    # Get total number of unique entities by type
    df_exploded = df.explode("db_ids").explode("type")
    ents_by_type = (
        df_exploded.groupby("type").agg({"db_ids": "nunique"}).to_dict()["db_ids"]
    )
    mention_counts_by_type = (
        df_exploded.groupby("type").agg({"db_ids": "count"}).to_dict()["db_ids"]
    )

    # Get distribution of entities
    if plot:
        test_df["in_train"] = test_df.db_ids_set.apply(lambda x: bool(x & unique_train))
        curie_counts = (
            test_df.explode("db_ids")
            .groupby("db_ids")
            .agg({"mention_id": "count", "in_train": "first"})
            .rename({"mention_id": "cui_num_test_mentions"}, axis=1)
        )
        sns.displot(
            data=curie_counts, x="cui_num_test_mentions", hue="in_train", log_scale=True
        )
        if dataset_name is not None:
            plt.title(dataset_name)
        plt.show()

    # Get overlap of mentions
    train_test_ent_overlap = len(unique_test & unique_train) / len(unique_test)
    test_ent_overlap = len(unique_test & unique_train_valid) / len(unique_test)
    mention_overlap = (
        test_df["db_ids_set"].apply(lambda x: bool(x & unique_train)).mean()
    )

    return {
        "unique_ents": len(unique_ents),
        "ent_overlap": train_test_ent_overlap,
        "mention_overlap": mention_overlap,
        "unique_types": len(unique_types),
        "total_documents": len(df.document_id.unique()),
        "train_documents": len(train_df.document_id.unique()),
        "test_documents": len(test_df.document_id.unique()),
        "total_mentions": df.shape[0],
        "train_mentions": train_df.shape[0],
        "test_mentions": test_df.shape[0],
        "has_validation_set": ("validation" in df.split.unique()),
    }



from bioel.models.arboel.biencoder.data.BiEncoderLightningDataModule import (
    ArboelDataModule,
)
from bioel.models.arboel.biencoder.model.BiEncoderLightningModule import LitArboel
from datetime import datetime
from lightning.pytorch.callbacks import ModelCheckpoint
from lightning.pytorch.loggers import WandbLogger
import wandb
from lightning.pytorch.utilities import rank_zero_only
import lightning.pytorch as L
from bioel.ontology import BiomedicalOntology
from bioel.models.arboel.biencoder.model.common.params import BlinkParser
import os
import pickle


@rank_zero_only
def create_ontology_object(args):
    if hasattr(BiomedicalOntology, args["load_function"]):
        load_func = getattr(BiomedicalOntology, args["load_function"])
        if args["ontology_dict"]:
            ontology_object = load_func(**args["ontology_dict"])
            print(f"Ontology loaded successfully. Name: {ontology_object.name}")
        else:
            raise ValueError("No ontology data provided.")
    else:
        raise ValueError(
            f"Error: {args['load_function']} is not a valid function for BiomedicalOntology."
        )

    file_path = os.path.join(args["data_path"], f"{args['ontology']}_object.pickle")
    os.makedirs(os.path.dirname(file_path), exist_ok=True)
    with open(file_path, "wb") as f:
        pickle.dump(ontology_object, f)
    return ontology_object


def train_model(params, model):
    print("params :", params)

    os.environ["WANDB_SILENT"] = "true"
    os.environ["NCCL_TIMEOUT"] = "10800"

    create_ontology_object(params)
    os.makedirs(params["output_path"], exist_ok=True)

    data_module = ArboelDataModule(params=params)

    current_time = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    model_checkpoint = ModelCheckpoint(
        monitor="max_acc",
        dirpath=params["output_path"],
        filename=f"biencoder_{current_time}-{{epoch}}-{{max_acc:.2f}}",
        save_top_k=1,
        mode="max",
        verbose=True,
    )

    if wandb.run is None:
        wandb.init()
    wandb_logger = WandbLogger(
        project=params["experiment"] if params["experiment"] else wandb.run.name
    )
    trainer = L.Trainer(
        num_sanity_val_steps=0,
        limit_train_batches=(
            params["limit_train_batches"] if params["limit_train_batches"] else 1.0
        ),
        limit_val_batches=1,
        max_epochs=params["num_train_epochs"],
        devices=params["devices"],
        accelerator="gpu" if params["devices"] else "cpu",
        strategy="ddp_find_unused_parameters_true",
        enable_progress_bar=True,
        callbacks=[model_checkpoint],
        accumulate_grad_batches=params["gradient_accumulation_steps"],
        precision="16-mixed",
        logger=wandb_logger,
    )

    trainer.fit(model, datamodule=data_module)



import sys

import os
import io
import json
import pickle
import logger
import numpy as np
import torch
import glob
import pandas as pd
from tqdm import tqdm

import lightning as L
from torch.utils.data import Dataset, DataLoader
from transformers import AutoTokenizer
from typing import Optional
from pytorch_metric_learning import samplers

from torch.utils.data import DataLoader

# import bioel.models.arboel.data.data_process as data_process
# from bioel.models.arboel.data.data_utils import process_mention_dataset
# from bioel.models.arboel.model.eval_cluster_linking import filter_by_context_doc_id
from bioel.ontology import BiomedicalOntology
from bioel.models.sapbert.data.utils import (
    generate_positive_pairs,
    MetricLearningDataset,
    MetricLearningDataset_pairwise,
    SapBertBigBioDataset,
    SapBertDictionaryDataset,
)

from bioel.utils import bigbio_utils

import logging

# Configure the logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class SapbertDataModule(L.LightningDataModule):
    """ """

    def __init__(self, params):
        """
        Parameters
        ----------
        - params : dict(str)
        Contains Configuration options
        """
        super().__init__()
        self.save_hyperparameters(params)

        self.data_path = self.hparams.train_dir
        self.tokenizer = AutoTokenizer.from_pretrained(self.hparams.model_dir)

        self.batch_size = self.hparams.train_batch_size

    def prepare_data(self):
        """
        Prepare the data
        """
        if self.hparams.mode == "pretrain":
            print(self.data_path)
            ontology_config = {
                "filepath": self.data_path,
                "name": "umls",
                "abbrev": None,
            }
            ontology = BiomedicalOntology.load_umls(**ontology_config)

            self.positive_pairs = generate_positive_pairs(ontology.entities)
        elif self.hparams.mode == "finetune":
            logger.info(f"Loading the BigBio Dataset {self.hparams.train_dir}")

            curie_dict = SapBertDictionaryDataset(self.hparams.train_dir).data
            self.positive_pairs = generate_positive_pairs(curie_dict)

        else:
            raise ValueError(f"Invalid mode: {self.hparams.mode}")

    def setup(self):
        """
        Setup the data
        """
        if self.hparams.pairwise:
            self.train_set = MetricLearningDataset_pairwise(
                training_pairs=self.positive_pairs, tokenizer=self.tokenizer
            )
        else:
            self.train_set = MetricLearningDataset(
                training_pairs=self.positive_pairs, tokenizer=self.tokenizer
            )

    def collate_fn_batch_encoding(self, batch):
        query1, query2, query_id = zip(*batch)
        query_encodings1 = self.tokenizer(
            list(query1),
            max_length=self.hparams.max_length,
            padding="max_length",
            add_special_tokens=True,
            truncation=True,
            return_tensors="pt",
        )
        query_encodings2 = self.tokenizer(
            list(query2),
            max_length=self.hparams.max_length,
            padding="max_length",
            add_special_tokens=True,
            truncation=True,
            return_tensors="pt",
        )
        query_ids = torch.tensor(list(query_id))
        return query_encodings1, query_encodings2, query_ids

    def train_dataloader(self):
        """
        Returns the training dataloader
        """
        if self.hparams.pairwise:
            print(len(self.train_set))
            return DataLoader(
                self.train_set,
                batch_size=self.batch_size,
                num_workers=self.hparams.num_workers,
                shuffle=True,
                collate_fn=self.collate_fn_batch_encoding,
            )
        else:
            return DataLoader(
                self.train_set,
                batch_size=self.batch_size,
                num_workers=self.hparams.num_workers,
                sampler=samplers.MPerClassSampler(
                    self.train_set.query_ids,
                    m=2,
                    length_before_new_iter=100000,
                ),
            )


if __name__ == "__main__":
    data_module = SapbertDataModule(
        params={
            "train_dir": "/mitchell/entity-linking/2017AA/META/",
            "model_dir": "microsoft/BiomedNLP-PubMedBERT-base-uncased-abstract-fulltext",
            "train_batch_size": 32,
            "num_workers": 4,
            "pairwise": True,
            "max_length": 128,
            "mode": "pretrain",
        }
    )
    data_module.prepare_data()
    data_module.setup()
    train_loader = data_module.train_dataloader()
    for batch in train_loader:
        print(batch)
        break



"""
Utilities for working with the local dataset cache.
"""

import os
import shutil
import tempfile
import json
from urllib.parse import urlparse
from pathlib import Path
from typing import Optional, Tuple, Union, IO
from hashlib import sha256

import requests
from tqdm import tqdm

CACHE_ROOT = Path(os.getenv("FILES_CACHE", "./Cached_files"))
DATASET_CACHE = str(CACHE_ROOT / "datasets")


def get_path(
    url_or_filename: Union[str, Path], cache_dir: Optional[str] = None
) -> str:
    """
    Given something that might be a URL (or might be a local path),
    determine which. If it's a URL, download the file and cache it, and
    return the path to the cached file. If it's already a local path,
    make sure the file exists and then return the path.
    """
    if cache_dir is None:
        cache_dir = DATASET_CACHE
    if isinstance(url_or_filename, Path):
        url_or_filename = str(url_or_filename)

    parsed = urlparse(url_or_filename)

    if parsed.scheme in ("http", "https"):
        # URL, so get it from the cache (downloading if necessary)
        return get_from_url(url_or_filename, cache_dir)
    elif os.path.exists(url_or_filename):
        # File, and it exists.
        return url_or_filename
    elif parsed.scheme == "":
        # File, but it doesn't exist.
        raise FileNotFoundError("file {} not found".format(url_or_filename))
    else:
        # Something unknown
        raise ValueError(
            "unable to parse {} as a URL or as a local path".format(url_or_filename)
        )


def url_to_filename(url: str, etag: Optional[str] = None) -> str:
    """
    Convert `url` into a hashed filename in a repeatable way.
    If `etag` is specified, append its hash to the url's, delimited
    by a period.
    """

    last_part = url.split("/")[-1]
    url_bytes = url.encode("utf-8")
    url_hash = sha256(url_bytes)
    filename = url_hash.hexdigest()

    if etag:
        etag_bytes = etag.encode("utf-8")
        etag_hash = sha256(etag_bytes)
        filename += "." + etag_hash.hexdigest()

    filename += "." + last_part
    return filename


def filename_to_url(filename: str, cache_dir: Optional[str] = None) -> Tuple[str, str]:
    """
    Return the url and etag (which may be ``None``) stored for `filename`.
    Raise ``FileNotFoundError`` if `filename` or its stored metadata do not exist.
    """
    if cache_dir is None:
        cache_dir = DATASET_CACHE

    cache_path = os.path.join(cache_dir, filename)
    if not os.path.exists(cache_path):
        raise FileNotFoundError("file {} not found".format(cache_path))

    meta_path = cache_path + ".json"
    if not os.path.exists(meta_path):
        raise FileNotFoundError("file {} not found".format(meta_path))

    with open(meta_path) as meta_file:
        metadata = json.load(meta_file)
    url = metadata["url"]
    etag = metadata["etag"]

    return url, etag


def http_get(url: str, temp_file: IO) -> None:
    req = requests.get(url, stream=True)
    total = int(req.headers.get("content-length", 0))
    pbar = tqdm(total=total, unit="iB", unit_scale=True, unit_divisor=1024)
    for chunk in req.iter_content(chunk_size=1024):
        if chunk:  # filter out keep-alive new chunks
            size = temp_file.write(chunk)
            pbar.update(size)
    pbar.close()


def get_from_url(url: str, cache_dir: Optional[str] = None) -> str:
    """
    Given a URL, look for the corresponding dataset in the local cache.
    If it's not there, download it. Then return the path to the cached file.
    """
    if cache_dir is None:
        cache_dir = DATASET_CACHE

    os.makedirs(cache_dir, exist_ok=True)

    response = requests.head(url, allow_redirects=True)
    if response.status_code != 200:
        raise IOError(
            "HEAD request failed for url {} with status code {}".format(
                url, response.status_code
            )
        )
    etag = response.headers.get("ETag")

    filename = url_to_filename(url, etag)

    # get cache path to put the file
    cache_path = os.path.join(cache_dir, filename)

    if not os.path.exists(cache_path):
        # Download to temporary file, then copy to cache dir once finished.
        # Otherwise you get corrupt cache entries if the download gets interrupted.
        with tempfile.NamedTemporaryFile() as temp_file:  # type: IO
            print(f"{url} not found in cache, downloading to {temp_file.name}")

            # GET file object
            http_get(url, temp_file)

            # we are copying the file before closing it, so flush to avoid truncation
            temp_file.flush()
            # shutil.copyfileobj() starts at the current position, so go to the start
            temp_file.seek(0)

            print(
                f"Finished download, copying {temp_file.name} to cache at {cache_path}"
            )
            with open(cache_path, "wb") as cache_file:
                shutil.copyfileobj(temp_file, cache_file)

            meta = {"url": url, "etag": etag}
            meta_path = cache_path + ".json"
            with open(meta_path, "w") as meta_file:
                json.dump(meta, meta_file)

    return cache_path


import spacy
from bioel.ontology import BiomedicalOntology
from bioel.models.scispacy.candidate_generation import CandidateGenerator
from bioel.models.scispacy.scispacy_embeddings import KnowledgeBaseEmbeddings
from bioel.models.scispacy.entity_linking import EntityLinker
from bioel.utils.bigbio_utils import (
    load_bigbio_dataset,
    add_deabbreviations,
    dataset_to_documents,
    dataset_to_df,
    load_dataset_df,
    resolve_abbreviation,
    CUIS_TO_REMAP,
    CUIS_TO_EXCLUDE,
    DATASET_NAMES,
    VALIDATION_DOCUMENT_IDS,
)
from bioel.utils.dataset_consts import (
    dataset_to_pretty_name,
    model_to_pretty_name,
    model_to_color,
)
import ujson
import os


def evaluate_model(
    dataset,
    ontology,
    k,
    path_to_save,
    output_path,
    equivalant_cuis=None,
    path_to_abbrev=None,
):
    """
    Params :
    ---------
    dataset: str
        Name of the dataset to evaluate
    ontology: BiomedicalOntology
        Ontology to use for candidate generation
    k: int
        Number of candidates to generate
    path_to_save: str
        Path to save the serialized knowledge base
    output_path: str
        Path to save the output
    equivalent_cuis: bool
        Whether the ontology has equivalent cuis or not
    path_to_abbrev: str
        Path to the abbreviations file
    """
    myembed = KnowledgeBaseEmbeddings(ontology)
    myembed.create_tfidf_ann_index(path_to_save)
    kb = myembed.serialized_kb()
    cand_gen = CandidateGenerator(kb)
    df = load_dataset_df(name=dataset, path_to_abbrev=path_to_abbrev)
    df = df.query("split == 'test'")

    output = list()

    if equivalant_cuis:
        cui_synsets = {}
        for cui, entity in ontology.entities.items():
            cui_synsets[cui] = entity.equivalant_cuis

    path_to_abbrev = True
    for index, row in df.iterrows():
        list_dbid = list(row.db_ids)  # Accessing db_ids as a list
        if path_to_abbrev:
            mention = row.deabbreviated_text  # Accessing the text of the current row
        else:
            mention = row.text
        candidates = cand_gen([mention], 3 * k)  # Generating candidates
        predicted = []

        for cand in candidates[0]:
            score = max(cand.similarities)
            predicted.append((cand.concept_id, score))

        sorted_predicted = sorted(predicted, reverse=True, key=lambda x: x[1])
        sorted_predicted = sorted_predicted[:k]  # Taking top k predictions

        cui_result = [[cui[0]] for cui in sorted_predicted]
        score = [[cui[1]] for cui in sorted_predicted]

        if equivalant_cuis:
            cui_result = [cui_synsets[y[0]] for y in cui_result]

        if path_to_abbrev:
            output.append(
                {
                    "document_id": row.document_id,
                    "offsets": row.offsets,
                    "text": row.text,
                    "type": row.type,
                    "db_ids": list_dbid,
                    "split": row.split,
                    "deabbreviated_text": row.deabbreviated_text,
                    "mention_id": row.mention_id + ".abbr_resolved",
                    "candidates": cui_result,
                    "scores": score,
                }
            )
        else:
            output.append(
                {
                    "document_id": row.document_id,
                    "offsets": row.offsets,
                    "text": row.text,
                    "type": row.type,
                    "db_ids": list_dbid,
                    "split": row.split,
                    "mention_id": row.mention_id,
                    "candidates": cui_result,
                    "scores": score,
                }
            )

    with open(output_path, "w") as f:
        f.write(ujson.dumps(output, indent=2))


if __name__ == "__main__":
    ontology_dir = "/mitchell/entity-linking/kbs/medic.tsv"
    myonto = BiomedicalOntology.load_medic(filepath=ontology_dir, name="medic")
    dataset = "ncbi_disease"

    # mesh_dict = {"name": "mesh", "filepath": "/mitchell/entity-linking/2017AA/META/"}
    # myonto = BiomedicalOntology.load_mesh(**mesh_dict)
    # dataset = "nlmchem"

    # ontology_dict = {
    #     "name": "entrez",
    #     "filepath": "/mitchell/entity-linking/el-robustness-comparison/data/gene_info.tsv",
    #     "dataset": "nlm_gene",
    # }
    # myonto = BiomedicalOntology.load_entrez(**ontology_dict)
    # dataset = "nlm_gene"

    # ontology_dict = {
    #     "name": "entrez",
    #     "filepath": "/mitchell/entity-linking/el-robustness-comparison/data/gene_info.tsv",
    #     "dataset": "gnormplus",
    # }
    # myonto = BiomedicalOntology.load_entrez(**ontology_dict)
    # dataset = "gnormplus"

    # umls_dict = {"name": "umls", "filepath": "/mitchell/entity-linking/2017AA/META/"}
    # myonto = BiomedicalOntology.load_umls(**umls_dict)
    # dataset = "medmentions_st21pv"

    # umls_dict = {"name": "umls", "filepath": "/mitchell/entity-linking/2017AA/META/"}
    # myonto = BiomedicalOntology.load_umls(**umls_dict)
    # dataset = "medmentions_full"

    path_to_save = f"/home2/cye73/data_test2/scispacy/kb_paths_scispacy/{dataset}"
    output_path = f"/home2/cye73/results2/scispacy/{dataset}_output.json"
    evaluate_model(
        dataset=dataset,
        ontology=myonto,
        k=10,
        path_to_save=path_to_save,
        output_path=output_path,
        path_to_abbrev="/home2/cye73/data_test2/abbreviations.json",
    )



# Copyright (c) Facebook, Inc. and its affiliates.
# Copyright (c) 2021 Dhruv Agarwal and authors of arboEL.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#
import collections
import os
import sys
import io
import torch
import logging
import random
import time
import numpy as np
import pickle
import json
from tqdm import tqdm, trange
from torch.utils.data import DataLoader, TensorDataset
from bioel.models.arboel.model.common.optimizer import get_bert_optimizer
from pytorch_transformers.optimization import WarmupLinearSchedule
from bioel.models.arboel.crossencoder.original.crossencoder import CrossEncoderRanker
from bioel.models.arboel.model.common.params import BlinkParser
import bioel.models.arboel.data.data_process as data_process
from IPython import embed

from bioel.logger import setup_logger

logger = setup_logger()

from pytorch_transformers.modeling_utils import CONFIG_NAME, WEIGHTS_NAME
OPTIM_SCHED_FNAME = "optim_sched.pth"


def read_dataset(dataset_name, preprocessed_json_data_parent_folder, debug=False):
    file_name = "{}.jsonl".format(dataset_name)
    txt_file_path = os.path.join(preprocessed_json_data_parent_folder, file_name)

    samples = []

    with io.open(txt_file_path, mode="r", encoding="utf-8") as file:
        for line in file:
            samples.append(json.loads(line.strip()))
            if debug and len(samples) > 200:
                break

    print(f"Read {len(samples)} samples.")
    return samples

def save_model(model, tokenizer, output_dir, scheduler=None, optimizer=None):
    """Saves the model and the tokenizer used in the output directory."""
    """Additionally, store the optimizer and scheduler to resume training."""
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    model_to_save = model.module if hasattr(model, "module") else model
    output_model_file = os.path.join(output_dir, WEIGHTS_NAME)
    output_config_file = os.path.join(output_dir, CONFIG_NAME)
    torch.save(model_to_save.state_dict(), output_model_file)
    model_to_save.config.to_json_file(output_config_file)
    tokenizer.save_vocabulary(output_dir)
    if scheduler is not None and optimizer is not None:
        optimizer_scheduler_file = os.path.join(output_dir, OPTIM_SCHED_FNAME)
        optim_sched = {
            'optimizer': optimizer,
            'scheduler': scheduler
        }
        torch.save(optim_sched, optimizer_scheduler_file)


def write_to_file(path, string, mode="w"):
    with open(path, mode) as writer:
        writer.write(string)


def accuracy(out, labels, return_bool_arr=False):
    outputs = np.argmax(out, axis=1)
    if return_bool_arr:
        return outputs == labels, outputs
    return np.sum(outputs == labels)


def modify(context_input, candidate_input, max_seq_length):
    new_input = []
    context_input = context_input.tolist()
    candidate_input = candidate_input.tolist()

    for i in tqdm(range(len(context_input)), desc="Concatenating"):
        cur_input = context_input[i]
        cur_candidate = candidate_input[i]
        mod_input = []
        for j in range(len(cur_candidate)):
            # remove [CLS] token from candidate
            sample = cur_input + cur_candidate[j][1:]
            sample = sample[:max_seq_length]
            mod_input.append(sample)

        new_input.append(mod_input)

    return torch.LongTensor(new_input)


def evaluate(
    reranker,
    eval_dataloader,
    device,
    logger,
    context_length,
    silent=True,
    unfiltered_length=None,
    mention_data=None,
    compute_macro_avg=False,
    store_failure_success=False,
):
    reranker.model.eval()
    if silent:
        iter_ = eval_dataloader
    else:
        iter_ = tqdm(eval_dataloader, desc="Evaluation")

    results = {}

    eval_accuracy = 0.0
    nb_eval_examples = 0
    nb_eval_steps = 0

    if mention_data is not None:
        processed_mention_data = mention_data["mention_data"]
        n_mentions_per_type = collections.defaultdict(int)
        if compute_macro_avg:
            for men in processed_mention_data:
                n_mentions_per_type[men["type"]] += 1
        dictionary = mention_data["entity_dictionary"]
        stored_candidates = mention_data["stored_candidates"]
        if store_failure_success:
            failsucc = {"failure": [], "success": []}
    n_evaluated_per_type = collections.defaultdict(int)
    n_hits_per_type = collections.defaultdict(int)

    for step, batch in enumerate(iter_):
        batch = tuple(t.to(device) for t in batch)
        context_input, label_input, mention_idxs = batch
        with torch.no_grad():
            eval_loss, logits = reranker(context_input, label_input, context_length)

        logits = logits.detach().cpu().numpy()
        label_ids = label_input.cpu().numpy()

        tmp_eval_hits, predicted = accuracy(logits, label_ids, return_bool_arr=True)
        tmp_eval_accuracy = np.sum(tmp_eval_hits)

        eval_accuracy += tmp_eval_accuracy

        nb_eval_examples += context_input.size(0)
        nb_eval_steps += 1

        if compute_macro_avg:
            for i, m_idx in enumerate(mention_idxs):
                type = processed_mention_data[m_idx]["type"]
                n_evaluated_per_type[type] += 1
                is_hit = tmp_eval_hits[i]
                n_hits_per_type[type] += is_hit

        if store_failure_success:
            for i, m_idx in enumerate(mention_idxs):
                m_idx = m_idx.item()
                men_query = processed_mention_data[m_idx]
                dict_pred = dictionary[
                    stored_candidates["candidates"][m_idx][predicted[i]]
                ]
                report_obj = {
                    "mention_id": men_query["mention_id"],
                    "mention_name": men_query["mention_name"],
                    "mention_gold_cui": "|".join(men_query["label_cuis"]),
                    "mention_gold_cui_name": "|".join(
                        [
                            dictionary[i]["title"]
                            for i in men_query["label_idxs"][: men_query["n_labels"]]
                        ]
                    ),
                    "predicted_name": dict_pred["title"],
                    "predicted_cui": dict_pred["cui"],
                }
                failsucc["success" if tmp_eval_hits[i] else "failure"].append(
                    report_obj
                )

    results["filtered_length"] = nb_eval_examples
    normalized_eval_accuracy = eval_accuracy / nb_eval_examples
    results["normalized_accuracy"] = normalized_eval_accuracy
    if unfiltered_length is not None:
        results["unfiltered_length"] = unfiltered_length
        results["unnormalized_accuracy"] = eval_accuracy / unfiltered_length
    if compute_macro_avg:
        norm_macro, unnorm_macro = 0, 0
        for type in n_mentions_per_type:
            norm_macro += n_hits_per_type[type] / n_evaluated_per_type[type]
            unnorm_macro += n_hits_per_type[type] / n_mentions_per_type[type]
        norm_macro /= len(n_evaluated_per_type)
        unnorm_macro /= len(n_mentions_per_type)
        results["normalized_macro_avg_acc"] = norm_macro
        results["unnormalized_macro_avg_acc"] = unnorm_macro
    if store_failure_success:
        results["failure"] = failsucc["failure"]
        results["success"] = failsucc["success"]
    if not store_failure_success:
        logger.info(json.dumps(results))
    return results


def load_data(
    data_split,
    bi_tokenizer,
    max_context_length,
    max_cand_length,
    knn,
    pickle_src_path,
    params,
    logger,
    return_dict_only=False,
):
    entity_dictionary_loaded = False
    entity_dictionary_pkl_path = os.path.join(
        pickle_src_path, "entity_dictionary.pickle"
    )
    if os.path.isfile(entity_dictionary_pkl_path):
        print("Loading stored processed entity dictionary...")
        with open(entity_dictionary_pkl_path, "rb") as read_handle:
            entity_dictionary = pickle.load(read_handle)
        entity_dictionary_loaded = True

    if return_dict_only and entity_dictionary_loaded:
        return entity_dictionary

    # Load data
    tensor_data_pkl_path = os.path.join(
        pickle_src_path, f"{data_split}_tensor_data.pickle"
    )
    processed_data_pkl_path = os.path.join(
        pickle_src_path, f"{data_split}_processed_data.pickle"
    )
    if os.path.isfile(tensor_data_pkl_path) and os.path.isfile(processed_data_pkl_path):
        print("Loading stored processed data...")
        with open(tensor_data_pkl_path, "rb") as read_handle:
            tensor_data = pickle.load(read_handle)
        with open(processed_data_pkl_path, "rb") as read_handle:
            processed_data = pickle.load(read_handle)
    else:
        data_samples = read_dataset(data_split, params["data_path"])
        if not entity_dictionary_loaded:
            with open(
                os.path.join(params["data_path"], "dictionary.pickle"), "rb"
            ) as read_handle:
                entity_dictionary = pickle.load(read_handle)

        # Check if dataset has multiple ground-truth labels
        mult_labels = "labels" in data_samples[0].keys()
        # Filter samples without gold entities
        data_samples = list(
            filter(
                lambda sample: (
                    (len(sample["labels"]) > 0)
                    if mult_labels
                    else (sample["label"] is not None)
                ),
                data_samples,
            )
        )
        logger.info("Read %d data samples." % len(data_samples))

        processed_data, entity_dictionary, tensor_data = (
            data_process.process_mention_data(
                data_samples,
                entity_dictionary,
                bi_tokenizer,
                max_context_length,
                max_cand_length,
                context_key=params["context_key"],
                multi_label_key="labels" if mult_labels else None,
                silent=params["silent"],
                logger=logger,
                debug=params["debug"],
                knn=knn,
                dictionary_processed=entity_dictionary_loaded,
            )
        )
        print("Saving processed data...")
        if not entity_dictionary_loaded:
            with open(entity_dictionary_pkl_path, "wb") as write_handle:
                pickle.dump(
                    entity_dictionary, write_handle, protocol=pickle.HIGHEST_PROTOCOL
                )
        with open(tensor_data_pkl_path, "wb") as write_handle:
            pickle.dump(tensor_data, write_handle, protocol=pickle.HIGHEST_PROTOCOL)
        with open(processed_data_pkl_path, "wb") as write_handle:
            pickle.dump(processed_data, write_handle, protocol=pickle.HIGHEST_PROTOCOL)

    if return_dict_only:
        return entity_dictionary
    return entity_dictionary, tensor_data, processed_data


def get_data_loader(
    data_split,
    tokenizer,
    context_length,
    candidate_length,
    max_seq_length,
    pickle_src_path,
    logger,
    inject_ground_truth=False,
    max_n=None,
    shuffle=True,
    return_data=False,
    custom_cand_set=None,
):
    # Load the top-64 indices for each mention query and the ground truth label if it exists in the candidate set
    logger.info(f"Loading {data_split} data...")
    cand_name = data_split
    if custom_cand_set is not None:
        logger.info(f"Loading custom candidate set: {custom_cand_set}...")
        cand_name = custom_cand_set
    fname = os.path.join(
        params["biencoder_indices_path"], f"candidates_{cand_name}_top64.t7"
    )
    stored_data = torch.load(fname)
    entity_dictionary, tensor_data, processed_data = load_data(
        data_split = data_split,
        tokenizer=tokenizer,
        context_length=context_length,
        candidate_length=candidate_length,
        knn=1,
        pickle_src_path=pickle_src_path,
        params = params,
        logger=logger,
    )
    logger.info("Loaded")
    dict_vecs = list(map(lambda x: x["ids"], entity_dictionary))

    mention_idxs = torch.tensor([i for i in range(len(stored_data["labels"]))])
    candidate_input = []
    keep_mask = [True] * len(stored_data["labels"])
    for i in tqdm(range(len(stored_data["labels"])), desc="Processing"):
        if stored_data["labels"][i] == -1:
            # If ground truth not in candidates, replace the last candidate with the ground truth
            if inject_ground_truth:
                gold_idx = processed_data[i]["label_idxs"][0]
                stored_data["labels"][i] = len(stored_data["candidates"][i]) - 1
                stored_data["candidates"][i][-1] = gold_idx
            else:
                keep_mask[i] = False
                continue
        cands = list(map(lambda x: dict_vecs[x], stored_data["candidates"][i]))
        candidate_input.append(cands)
    candidate_input = np.array(candidate_input)
    context_input = tensor_data[:][0][keep_mask]
    label_input = torch.tensor(stored_data["labels"])[keep_mask]
    mention_idxs = mention_idxs[keep_mask]

    n_no_label = len(stored_data["labels"]) - np.sum(keep_mask)

    if max_n is not None:
        context_input = context_input[:max_n]
        candidate_input = candidate_input[:max_n]
        label_input = label_input[:max_n]

    if params["debug"]:
        max_n = 200
        context_input = context_input[:max_n]
        candidate_input = candidate_input[:max_n]
        label_input = label_input[:max_n]

    context_input = modify(context_input, candidate_input, max_seq_length)
    tensor_data = TensorDataset(context_input, label_input, mention_idxs)
    dataloader = DataLoader(
        tensor_data,
        shuffle=shuffle,
        batch_size=params[
            "train_batch_size" if data_split == "train" else "eval_batch_size"
        ],
    )
    if return_data:
        return (
            dataloader,
            n_no_label,
            {
                "entity_dictionary": entity_dictionary,
                "mention_data": processed_data,
                "stored_candidates": stored_data,
            },
        )
    return dataloader, n_no_label





def main(params):
    model_output_path = params["output_path"]
    if not os.path.exists(model_output_path):
        os.makedirs(model_output_path)
    pickle_src_path = params["pickle_src_path"]
    logger = setup_logger()

    # Init model
    reranker = CrossEncoderRanker(params)
    tokenizer = reranker.tokenizer
    model = reranker.model
    device = reranker.device
    n_gpu = reranker.n_gpu

    if params["gradient_accumulation_steps"] < 1:
        raise ValueError(
            "Invalid gradient_accumulation_steps parameter: {}, should be >= 1".format(
                params["gradient_accumulation_steps"]
            )
        )
    # An effective batch size of `x`, when we are accumulating the gradient accross `y` batches will be achieved by having a batch size of `z = x / y`
    # args.gradient_accumulation_steps = args.gradient_accumulation_steps // n_gpu
    params["train_batch_size"] = (
        params["train_batch_size"] // params["gradient_accumulation_steps"]
    )
    grad_acc_steps = params["gradient_accumulation_steps"]

    # Fix the random seeds
    seed = params["seed"]
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if reranker.n_gpu > 0:
        torch.cuda.manual_seed_all(seed)

    max_seq_length = params["max_seq_length"]
    context_length = params["max_context_length"]
    candidate_length = params["max_context_length"]

    if params["only_evaluate"]:
        test_dataloader, n_test_skipped, data = get_data_loader(
            "test",
            tokenizer,
            context_length,
            candidate_length,
            max_seq_length,
            pickle_src_path,
            logger,
            inject_ground_truth=params["inject_eval_ground_truth"],
            shuffle=False,
            return_data=True,
            custom_cand_set=params["custom_cand_set"],
        )
        logger.info("Evaluating the model on the test set")
        results = evaluate(
            reranker,
            test_dataloader,
            device=device,
            logger=logger,
            context_length=context_length,
            silent=params["silent"],
            unfiltered_length=len(data["mention_data"]),
            mention_data=data,
            compute_macro_avg=True,
            store_failure_success=True,
        )
        results_path = os.path.join(model_output_path, "results.json")
        with open(results_path, "w") as f:
            json.dump(results, f, indent=2)
            print(f"\nAnalysis saved at: {results_path}")
        exit()

    train_dataloader, _, train_data = get_data_loader(
        "train",
        tokenizer,
        context_length,
        candidate_length,
        max_seq_length,
        pickle_src_path,
        logger,
        inject_ground_truth=params["inject_train_ground_truth"],
        return_data=True,
    )

    valid_dataloader, n_valid_skipped = get_data_loader(
        "valid",
        tokenizer,
        context_length,
        candidate_length,
        max_seq_length,
        pickle_src_path,
        logger,
        inject_ground_truth=params["inject_eval_ground_truth"],
        max_n=2048,
    )

    if not params["skip_initial_eval"]:
        logger.info("Evaluating dev set on untrained model...")
        # Evaluate before training
        results = evaluate(
            reranker,
            valid_dataloader,
            device=device,
            logger=logger,
            context_length=context_length,
            silent=params["silent"],
        )

    time_start = time.time()

    write_to_file(os.path.join(model_output_path, "training_params.txt"), str(params))

    logger.info("Starting training")
    logger.info(
        "device: {} n_gpu: {}, distributed training: {}".format(device, n_gpu, False)
    )

    optimizer = get_optimizer(model, params)
    scheduler = get_scheduler(
        params, optimizer, len(train_data["mention_data"]), logger
    )

    model.train()

    best_epoch_idx = -1
    best_score = -1

    num_train_epochs = params["num_train_epochs"]

    for epoch_idx in trange(int(num_train_epochs), desc="Epoch"):
        tr_loss = 0
        results = None

        if params["silent"]:
            iter_ = train_dataloader
        else:
            iter_ = tqdm(train_dataloader, desc="Batch")

        part = 0
        for step, batch in enumerate(iter_):
            batch = tuple(t.to(device) for t in batch)
            context_input, label_input, _ = batch
            loss, _ = reranker(context_input, label_input, context_length)

            if grad_acc_steps > 1:
                loss = loss / grad_acc_steps

            tr_loss += loss.item()

            if (step + 1) % (params["print_interval"] * grad_acc_steps) == 0:
                logger.info(
                    "Step {} - epoch {} average loss: {}\n".format(
                        step,
                        epoch_idx,
                        tr_loss / (params["print_interval"] * grad_acc_steps),
                    )
                )
                tr_loss = 0

            loss.backward()

            if (step + 1) % grad_acc_steps == 0:
                torch.nn.utils.clip_grad_norm_(
                    model.parameters(), params["max_grad_norm"]
                )
                optimizer.step()
                scheduler.step()
                optimizer.zero_grad()

            if params["eval_interval"] != -1:
                if (step + 1) % (params["eval_interval"] * grad_acc_steps) == 0:
                    logger.info("Evaluation on the development dataset")
                    evaluate(
                        reranker,
                        valid_dataloader,
                        device=device,
                        logger=logger,
                        context_length=context_length,
                        silent=params["silent"],
                    )
                    logger.info("***** Saving fine - tuned model *****")
                    epoch_output_folder_path = os.path.join(
                        model_output_path, "epoch_{}_{}".format(epoch_idx, part)
                    )
                    part += 1
                    save_model(model, tokenizer, epoch_output_folder_path)
                    model.train()
                    logger.info("\n")

        logger.info("***** Saving fine - tuned model *****")
        epoch_output_folder_path = os.path.join(
            model_output_path, "epoch_{}".format(epoch_idx)
        )
        save_model(model, tokenizer, epoch_output_folder_path)

        results = evaluate(
            reranker,
            valid_dataloader,
            device=device,
            logger=logger,
            context_length=context_length,
            silent=params["silent"],
        )

        ls = [best_score, results["normalized_accuracy"]]
        li = [best_epoch_idx, epoch_idx]

        best_score = ls[np.argmax(ls)]
        best_epoch_idx = li[np.argmax(ls)]
        logger.info("\n")

    execution_time = (time.time() - time_start) / 60
    write_to_file(
        os.path.join(model_output_path, "training_time.txt"),
        "The training took {} minutes\n".format(execution_time),
    )
    logger.info("The training took {} minutes\n".format(execution_time))

    # save the best model in the parent_dir
    logger.info("Best performance in epoch: {}".format(best_epoch_idx))


if __name__ == "__main__":
    parser = BlinkParser(add_model_args=True)
    parser.add_training_args()
    args = parser.parse_args()
    print(args)
    params = args.__dict__
    main(params)



import torch


def reform_input(inputs, attention_mask = None, ending_token = 2):
    
    ## input a tensor of size BSZ x Length
    max_idx = torch.max(torch.where(inputs==ending_token)[1])
    inputs = inputs[:, :max_idx+1]
    if attention_mask is not None:
        attention_mask = attention_mask[:, :max_idx+1]

    return inputs, attention_mask


from setuptools import setup, find_packages, Extension
import os
import subprocess

try:
    from Cython.Build import cythonize
except ImportError:
    subprocess.check_call([os.sys.executable, "-m", "pip", "install", "cython"])
    from Cython.Build import cythonize

try:
    import numpy
except ImportError:
    subprocess.check_call([os.sys.executable, "-m", "pip", "install", "numpy"])
    import numpy

# Define the Cython extension
cython_extensions = [
    Extension(
        "bioel.models.arboel.biencoder.model.special_partition.special_partition",
        ["bioel/models/arboel/biencoder/model/special_partition/special_partition.pyx"],
        include_dirs=[numpy.get_include()],
        define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
    )
]

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

# Setup function to include the Cython extension
setup(
    name="bioel",
    version="0.1.3",
    description="An easy-to-use package for all your biomedical entity linking needs.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/davidkartchner/biomedical-entity-linking/tree/main/bioel",
    license="MIT",
    license_files=["LICENSE"],
    author="Pathology Dynamics Lab, Georgia Institute of Technology",
    author_email="",
    keywords=[
        "bioel",
        "biomedical",
        "entity",
        "linking",
        "entity-linking",
        "biomedical-entity-linking",
    ],
    packages=find_packages(),
    python_requires=">=3.9",
    setup_requires=["numpy", "cython"],  # Ensures numpy and cython are installed early
    install_requires=[
        "pytest",
        "tqdm",
        "pandas",
        "numpy",
        "matplotlib",
        "seaborn",
        "obonet",
        "datasets",
        "ujson",
        "bioc",
        "logger",
        "lightning",
        "ipython",
        "cython",
        "transformers==4.44.2",
        "pytorch_transformers",
        "scipy",
        "faiss-gpu",
        "faiss-cpu",
        "wandb",
        "scikit-learn",
        "torch",
        "scispacy",
        # "fairseq = 0.12.2",
        "accelerate",
        "pytorch_metric_learning",
    ],
    ext_modules=cythonize(cython_extensions),
    include_package_data=True,
    zip_safe=False,
)



# Copyright (c) Facebook, Inc. and its affiliates.
# Copyright (c) 2021 Dhruv Agarwal and authors of arboEL.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#
import os
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from collections import OrderedDict
from tqdm import tqdm
from pytorch_transformers.modeling_utils import CONFIG_NAME, WEIGHTS_NAME

from pytorch_transformers.modeling_bert import (
    BertPreTrainedModel,
    BertConfig,
    BertModel,
)

from pytorch_transformers.modeling_roberta import (
    RobertaConfig,
    RobertaModel,
)

from pytorch_transformers.tokenization_bert import BertTokenizer
from pytorch_transformers.tokenization_roberta import RobertaTokenizer

from bioel.models.arboel.model.common.ranker_base import BertEncoder, get_model_obj
from bioel.models.arboel.model.common.optimizer import get_bert_optimizer
from bioel.models.arboel.model.common.params import (
    ENT_START_TAG,
    ENT_END_TAG,
    ENT_TITLE_TAG,
)

from IPython import embed


def load_crossencoder(params):
    # Init model
    crossencoder = CrossEncoderRanker(params)
    return crossencoder


class CrossEncoderModule(torch.nn.Module):
    def __init__(self, params, tokenizer, start_mention_id=None, end_mention_id=None):
        super(CrossEncoderModule, self).__init__()
        model_path = params["bert_model"]
        if params.get("roberta"):
            ctxt_bert = RobertaModel.from_pretrained(model_path)
            cand_bert = RobertaModel.from_pretrained(model_path)
        else:
            ctxt_bert = BertModel.from_pretrained(model_path)
            cand_bert = BertModel.from_pretrained(model_path)
        ctxt_bert.resize_token_embeddings(len(tokenizer))
        cand_bert.resize_token_embeddings(len(tokenizer))
        self.pool_highlighted = params["pool_highlighted"]
        self.ctxt_encoder = BertEncoder(
            ctxt_bert,
            params["out_dim"],
            layer_pulled=params["pull_from_layer"],
            add_linear=params["add_linear"] and not self.pool_highlighted,
            get_all_outputs=self.pool_highlighted,
        )
        self.cand_encoder = BertEncoder(
            cand_bert,
            params["out_dim"],
            layer_pulled=params["pull_from_layer"],
            add_linear=params["add_linear"],
            get_all_outputs=False,
        )
        self.config = self.ctxt_encoder.bert_model.config
        self.start_mention_id = start_mention_id
        self.end_mention_id = end_mention_id

        if self.pool_highlighted:
            bert_output_dim = ctxt_bert.embeddings.word_embeddings.weight.size(1)
            output_dim = params["out_dim"]
            self.additional_linear = nn.Linear(2 * bert_output_dim, output_dim)
            self.dropout = nn.Dropout(0.1)

    def forward(self, token_idx_ctxt, segment_idx_ctxt, mask_ctxt, is_ctxt=True):
        if is_ctxt and self.pool_highlighted:
            with torch.no_grad():
                index_tensor = torch.arange(token_idx_ctxt.shape[1])
                index_tensor = index_tensor.repeat(token_idx_ctxt.shape[0], 1)
                index_tensor = index_tensor.to(token_idx_ctxt.device)

                start_indices = torch.nonzero(
                    token_idx_ctxt == self.start_mention_id, as_tuple=False
                )
                start_indices = start_indices[:, 1:]
                end_indices = torch.nonzero(
                    token_idx_ctxt == self.end_mention_id, as_tuple=False
                )
                end_indices = end_indices[:, 1:]

                start_indices_a = start_indices[::2, :]
                end_indices_a = end_indices[::2, :]
                start_indices_b = start_indices[1::2, :]
                end_indices_b = end_indices[1::2, :]

                mask_a = (index_tensor > start_indices_a) & (
                    index_tensor < end_indices_a
                )
                mask_b = (index_tensor > start_indices_b) & (
                    index_tensor < end_indices_b
                )
                mask_a = mask_a.unsqueeze(-1).type(torch.float)
                mask_b = mask_b.unsqueeze(-1).type(torch.float)

        embedding_ctxt = (
            self.ctxt_encoder(token_idx_ctxt, segment_idx_ctxt, mask_ctxt)
            if is_ctxt
            else self.cand_encoder(token_idx_ctxt, segment_idx_ctxt, mask_ctxt)
        )

        if is_ctxt and self.pool_highlighted:
            pooled_output_a = torch.sum(embedding_ctxt * mask_a, 1) / torch.sum(
                mask_a, 1
            )
            pooled_output_b = torch.sum(embedding_ctxt * mask_b, 1) / torch.sum(
                mask_b, 1
            )
            pooled_output = torch.cat((pooled_output_a, pooled_output_b), dim=1)
            embedding_ctxt = self.additional_linear(self.dropout(pooled_output))

        return embedding_ctxt.squeeze(-1)


class CrossEncoderRanker(torch.nn.Module):
    def __init__(self, params):
        super(CrossEncoderRanker, self).__init__()
        self.params = params
        self.device = torch.device(
            "cuda" if torch.cuda.is_available() and not params["no_cuda"] else "cpu"
        )
        self.n_gpu = torch.cuda.device_count()

        if params.get("roberta"):
            self.tokenizer = RobertaTokenizer.from_pretrained(
                params["bert_model"], do_lower_case=params["lowercase"]
            )
        else:
            self.tokenizer = BertTokenizer.from_pretrained(
                params["bert_model"], do_lower_case=params["lowercase"]
            )

        special_tokens_dict = {
            "additional_special_tokens": [
                ENT_START_TAG,
                ENT_END_TAG,
                ENT_TITLE_TAG,
            ],
        }
        self.tokenizer.add_special_tokens(special_tokens_dict)
        self.NULL_IDX = self.tokenizer.pad_token_id
        self.START_TOKEN = self.tokenizer.cls_token
        self.END_TOKEN = self.tokenizer.sep_token
        self.START_MENTION_ID = self.tokenizer.convert_tokens_to_ids(ENT_START_TAG)
        self.END_MENTION_ID = self.tokenizer.convert_tokens_to_ids(ENT_END_TAG)

        # keep some parameters around
        self.add_sigmoid = params["add_sigmoid"]
        self.margin = params["margin"]
        self.objective = params["objective"]
        self.pos_neg_loss = params.get("pos_neg_loss", False)
        assert self.objective == "softmax" or self.objective == "max_margin"

        # init model
        self.build_model()
        if params["path_to_model"] is not None:
            self.load_model(params["path_to_model"])

        self.model = self.model.to(self.device)
        self.data_parallel = params.get("data_parallel")
        if self.data_parallel:
            self.model = torch.nn.DataParallel(self.model)

    def load_model(self, fname, cpu=False):
        if cpu:
            state_dict = torch.load(fname, map_location=lambda storage, location: "cpu")
        else:
            state_dict = torch.load(fname)
        self.model.load_state_dict(state_dict)

    def save(self, output_dir):
        self.save_model(output_dir)
        self.tokenizer.save_vocabulary(output_dir)

    def build_model(self):
        self.model = CrossEncoderModule(
            self.params,
            self.tokenizer,
            start_mention_id=self.START_MENTION_ID,
            end_mention_id=self.END_MENTION_ID,
        )

    def save_model(self, output_dir):
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        model_to_save = get_model_obj(self.model)
        output_model_file = os.path.join(output_dir, WEIGHTS_NAME)
        output_config_file = os.path.join(output_dir, CONFIG_NAME)
        torch.save(model_to_save.state_dict(), output_model_file)
        model_to_save.config.to_json_file(output_config_file)

    def get_optimizer(self):
        return get_bert_optimizer(
            [self.model],
            self.params["type_optimization"],
            self.params["learning_rate"],
            fp16=self.params.get("fp16"),
        )

    def score_candidate(
        self, text_vecs, context_len, is_context_encoder=True, no_sigmoid=False
    ):
        # Encode contexts first
        num_cand = text_vecs.size(1)
        text_vecs = text_vecs.reshape(-1, text_vecs.size(-1))
        token_idx_ctxt, segment_idx_ctxt, mask_ctxt = to_bert_input(
            text_vecs,
            self.NULL_IDX,
            context_len,
        )
        embedding_ctxt = self.model(
            token_idx_ctxt, segment_idx_ctxt, mask_ctxt, is_ctxt=is_context_encoder
        )
        if not no_sigmoid and self.add_sigmoid:
            embedding_ctxt = torch.sigmoid(embedding_ctxt)
        return embedding_ctxt.view(-1, num_cand)

    def forward(
        self, pos_scores, neg_ctxt_vecs, neg_cand_vecs, context_len, no_sigmoid=False
    ):
        n_negs = neg_cand_vecs.size(1) + (
            0 if neg_ctxt_vecs is None else neg_ctxt_vecs.size(1)
        )
        labels = torch.tensor(
            [[1] + [0] * n_negs] * len(pos_scores), dtype=torch.float32
        ).cuda()  # Shape: B x (1+knn_negs)
        cand_scores = self.score_candidate(
            neg_cand_vecs, context_len, is_context_encoder=False, no_sigmoid=no_sigmoid
        )  # Shape: B x knn_cand_negs
        scores = torch.cat(
            (pos_scores, cand_scores), dim=1
        )  # Shape: B x (1+knn_cand_negs)
        if neg_ctxt_vecs is not None:
            ctxt_scores = self.score_candidate(
                neg_ctxt_vecs,
                context_len,
                is_context_encoder=True,
                no_sigmoid=no_sigmoid,
            )  # Shape: B x knn_ctxt_negs
            scores = torch.cat(
                (scores, ctxt_scores), dim=1
            )  # Shape: B x (1+knn_cand_negs+knn_ctxt_negs)
        if self.pos_neg_loss:
            loss = torch.mean(
                torch.sum(
                    -torch.log(torch.softmax(scores, dim=1) + 1e-8) * labels
                    - torch.log(1 - torch.softmax(scores, dim=1) + 1e-8) * (1 - labels),
                    dim=1,
                )
            )
        else:
            loss = torch.mean(
                torch.max(
                    -torch.log(torch.softmax(scores, dim=1) + 1e-8) * labels, dim=1
                )[0]
            )
        return loss


def to_bert_input(token_idx, null_idx, segment_pos):
    """token_idx is a 2D tensor int.
    return token_idx, segment_idx and mask
    """
    segment_idx = token_idx * 0
    if segment_pos > 0:
        segment_idx[:, segment_pos:] = token_idx[:, segment_pos:] > 0

    mask = token_idx != null_idx
    # nullify elements in case self.NULL_IDX was not 0
    # token_idx = token_idx * mask.long()
    return token_idx, segment_idx, mask



# Copyright (c) Facebook, Inc. and its affiliates.
# Copyright (c) 2021 Dhruv Agarwal and authors of arboEL.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import os
import json
import random
import math
import time
import torch
import numpy as np
from tqdm import tqdm
import pickle
import copy

import bioel.models.arboel.biencoder.data.data_process as data_process
from bioel.models.arboel.biencoder.model.biencoder import BiEncoderRanker
from bioel.models.arboel.crossencoder.model.crossencoder import CrossEncoderRanker
from bioel.models.arboel.biencoder.model.common.params import BlinkParser
from bioel.models.arboel.biencoder.data.BiEncoderLightningDataModule import (
    read_dataset,
)

from IPython import embed

from bioel.logger import setup_logger


def load_data(
    data_split,
    bi_tokenizer,
    knn,
    params,
    logger,
    return_dict_only=False,
):
    entity_dictionary_loaded = False
    entity_dictionary_pkl_path = os.path.join(
        params["data_path"], "entity_dictionary.pickle"
    )
    if os.path.isfile(entity_dictionary_pkl_path):
        print("Loading stored processed entity dictionary...")
        with open(entity_dictionary_pkl_path, "rb") as read_handle:
            entity_dictionary = pickle.load(read_handle)
        entity_dictionary_loaded = True

    if return_dict_only and entity_dictionary_loaded:
        return entity_dictionary

    # Load data
    tensor_data_pkl_path = os.path.join(
        params["data_path"], f"{data_split}_tensor_data.pickle"
    )
    processed_data_pkl_path = os.path.join(
        params["data_path"], f"{data_split}_processed_data.pickle"
    )
    if os.path.isfile(tensor_data_pkl_path) and os.path.isfile(processed_data_pkl_path):
        print("Loading stored processed data...")
        with open(tensor_data_pkl_path, "rb") as read_handle:
            tensor_data = pickle.load(read_handle)
        with open(processed_data_pkl_path, "rb") as read_handle:
            processed_data = pickle.load(read_handle)
    else:
        data_samples = read_dataset(data_split, params["data_path"])
        if not entity_dictionary_loaded:
            with open(
                os.path.join(params["data_path"], "dictionary.pickle"), "rb"
            ) as read_handle:
                entity_dictionary = pickle.load(read_handle)

        # Check if dataset has multiple ground-truth labels
        mult_labels = "labels" in data_samples[0].keys()
        # Filter samples without gold entities
        data_samples = list(
            filter(
                lambda sample: (
                    (len(sample["labels"]) > 0)
                    if mult_labels
                    else (sample["label"] is not None)
                ),
                data_samples,
            )
        )
        logger.info("Read %d data samples." % len(data_samples))

        processed_data, entity_dictionary, tensor_data = (
            data_process.process_mention_data(
                data_samples,
                entity_dictionary,
                bi_tokenizer,
                max_context_length=params["max_context_length"],
                max_cand_length=params["max_cand_length"],
                context_key=params["context_key"],
                multi_label_key="labels" if mult_labels else None,
                silent=params["silent"],
                logger=logger,
                debug=params["debug"],
                knn=knn,
                dictionary_processed=entity_dictionary_loaded,
            )
        )
        print("Saving processed data...")
        if not entity_dictionary_loaded:
            with open(entity_dictionary_pkl_path, "wb") as write_handle:
                pickle.dump(
                    entity_dictionary, write_handle, protocol=pickle.HIGHEST_PROTOCOL
                )
        with open(tensor_data_pkl_path, "wb") as write_handle:
            pickle.dump(tensor_data, write_handle, protocol=pickle.HIGHEST_PROTOCOL)
        with open(processed_data_pkl_path, "wb") as write_handle:
            pickle.dump(processed_data, write_handle, protocol=pickle.HIGHEST_PROTOCOL)

    if return_dict_only:
        return entity_dictionary
    return entity_dictionary, tensor_data, processed_data


def save_topk_biencoder_cands(
    bi_reranker,
    logger,
    params,
    bi_tokenizer,
    topk=64,
):
    entity_dictionary = load_data(
        data_split="train",
        bi_tokenizer=bi_tokenizer,
        knn=1,
        params=params,
        logger=logger,
        return_dict_only=True,
    )
    entity_dict_vecs = torch.tensor(
        list(map(lambda x: x["ids"], entity_dictionary)), dtype=torch.long
    )

    logger.info("Biencoder: Embedding and indexing entity dictionary")
    if params["use_types"]:
        _, dict_indexes, dict_idxs_by_type = data_process.embed_and_index(
            bi_reranker,
            entity_dict_vecs,
            encoder_type="candidate",
            corpus=entity_dictionary,
            force_exact_search=True,
            batch_size=params["embed_batch_size"],
        )
    else:
        _, dict_index = data_process.embed_and_index(
            bi_reranker,
            entity_dict_vecs,
            encoder_type="candidate",
            force_exact_search=True,
            batch_size=params["embed_batch_size"],
        )
    logger.info("Biencoder: Embedding and indexing finished")

    print("NCCL_TIMEOUT:", os.environ.get("NCCL_TIMEOUT"))

    for mode in ["train", "valid", "test"]:
        logger.info(
            f"Biencoder: Fetching top-{topk} biencoder candidates for {mode} set"
        )
        _, tensor_data, processed_data = load_data(
            data_split=mode,
            bi_tokenizer=bi_tokenizer,
            knn=1,
            params=params,
            logger=logger,
        )
        men_vecs = tensor_data[:][0]

        logger.info("Biencoder: Embedding mention data")
        if params["use_types"]:
            men_embeddings, _, men_idxs_by_type = data_process.embed_and_index(
                bi_reranker,
                men_vecs,
                encoder_type="context",
                corpus=processed_data,
                force_exact_search=True,
                batch_size=params["embed_batch_size"],
            )
        else:
            men_embeddings = data_process.embed_and_index(
                bi_reranker,
                men_vecs,
                encoder_type="context",
                force_exact_search=True,
                batch_size=params["embed_batch_size"],
                only_embed=True,
            )
        logger.info("Biencoder: Embedding finished")

        logger.info("Biencoder: Finding nearest entities for each mention...")
        if not params["use_types"]:
            _, bi_dict_nns = dict_index.search(men_embeddings, topk)
        else:
            bi_dict_nns = np.zeros((len(men_embeddings), topk), dtype=int)
            for entity_type in men_idxs_by_type:
                men_embeds_by_type = men_embeddings[men_idxs_by_type[entity_type]]
                _, dict_nns_by_type = dict_indexes[entity_type].search(
                    men_embeds_by_type, topk
                )
                dict_nns_idxs = np.array(
                    list(
                        map(
                            lambda x: dict_idxs_by_type[entity_type][x],
                            dict_nns_by_type,
                        )
                    )
                )
                for i, idx in enumerate(men_idxs_by_type[entity_type]):
                    bi_dict_nns[idx] = dict_nns_idxs[i]
        logger.info("Biencoder: Search finished")

        labels = [-1] * len(bi_dict_nns)
        for men_idx in range(len(bi_dict_nns)):
            gold_idx = processed_data[men_idx]["label_idxs"][0]
            for i in range(len(bi_dict_nns[men_idx])):
                if bi_dict_nns[men_idx][i] == gold_idx:
                    labels[men_idx] = i
                    break

        logger.info(f"Biencoder: Saving top-{topk} biencoder candidates for {mode} set")
        save_data_path = os.path.join(
            params["data_path"], f"candidates_{mode}_top{topk}.t7"
        )
        torch.save(
            {"mode": mode, "candidates": bi_dict_nns, "labels": labels}, save_data_path
        )
        logger.info("Biencoder: Saved")

    return {"mode": mode, "candidates": bi_dict_nns, "labels": labels}
